"""
This module provides a convenience calculator for computing a single-center equivariant
power spectrum.
"""

import json
from typing import List, Optional, Union

from . import _dispatch
from ._backend import (
    CalculatorBase,
    Device,
    DType,
    IntoSystem,
    Labels,
    TensorMap,
    TorchModule,
    operations,
)
from ._cg_product import ClebschGordanProduct
from ._density_correlations import _filter_redundant_keys


class EquivariantPowerSpectrum(TorchModule):
    r"""
    Computes a general equivariant power spectrum descriptor of two calculators.

    If only ``calculator_1`` is provided, the power spectrum is computed as the density
    auto-correlation of the density produced by the first calculator. If
    ``calculator_2`` is also provided, the power spectrum is computed as the density
    cross-correlation of the densities produced by the two calculators.

    Example
    -------
    As an example we calculate the equivariant power spectrum for a short range (sr)
    spherical expansion and a long-range (lr) LODE spherical expansion for a NaCl
    crystal.

    >>> import featomic
    >>> import ase

    Construct the NaCl crystal

    >>> atoms = ase.Atoms(
    ...     symbols="NaCl",
    ...     positions=[[0, 0, 0], [0.5, 0.5, 0.5]],
    ...     pbc=True,
    ...     cell=[1, 1, 1],
    ... )

    Define the hyper parameters for the short-range spherical expansion

    >>> sr_hypers = {
    ...     "cutoff": {
    ...         "radius": 1.0,
    ...         "smoothing": {"type": "ShiftedCosine", "width": 0.5},
    ...     },
    ...     "density": {
    ...         "type": "Gaussian",
    ...         "width": 0.3,
    ...     },
    ...     "basis": {
    ...         "type": "TensorProduct",
    ...         "max_angular": 2,
    ...         "radial": {"type": "Gto", "max_radial": 5},
    ...     },
    ... }

    Define the hyper parameters for the long-range LODE spherical expansion from the
    hyper parameters of the short-range spherical expansion

    >>> lr_hypers = {
    ...     "density": {
    ...         "type": "SmearedPowerLaw",
    ...         "smearing": 0.3,
    ...         "exponent": 1,
    ...     },
    ...     "basis": {
    ...         "type": "TensorProduct",
    ...         "max_angular": 2,
    ...         "radial": {"type": "Gto", "max_radial": 3, "radius": 1.0},
    ...     },
    ... }

    Construct the calculators

    >>> sr_calculator = featomic.SphericalExpansion(**sr_hypers)
    >>> lr_calculator = featomic.LodeSphericalExpansion(**lr_hypers)

    Construct the power spectrum calculators and compute the spherical expansion

    >>> calculator = featomic.clebsch_gordan.EquivariantPowerSpectrum(
    ...     sr_calculator, lr_calculator
    ... )
    >>> power_spectrum = calculator.compute(atoms, neighbors_to_properties=True)

    The resulting equivariants are stored as :py:class:`metatensor.TensorMap` as for any
    other calculator. The keys contain the symmetry information:

    >>> power_spectrum.keys
    Labels(
        o3_lambda  o3_sigma  center_type
            0         1          11
            1         1          11
            2         1          11
            1         -1         11
            2         -1         11
            3         1          11
            3         -1         11
            4         1          11
            0         1          17
            1         1          17
            2         1          17
            1         -1         17
            2         -1         17
            3         1          17
            3         -1         17
            4         1          17
    )

    The block properties contain the angular order of the combined blocks ("l_1",
    "l_2"), along with the neighbor types ("neighbor_1_type", "neighbor_2_type") and
    radial channel indices.

    >>> power_spectrum[0].properties.names
    ['l_1', 'l_2', 'neighbor_1_type', 'n_1', 'neighbor_2_type', 'n_2']

    .. seealso::
        Faster power spectrum calculator specifically for invariant descriptors can
        be found at :py:class:`featomic.SoapPowerSpectrum` and
        :py:class:`featomic.clebsch_gordan.PowerSpectrum`.
    """

    def __init__(
        self,
        calculator_1: CalculatorBase,
        calculator_2: Optional[CalculatorBase] = None,
        neighbor_types: Optional[List[int]] = None,
        *,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
    ):
        """
        Constructs the equivariant power spectrum calculator.

        :param calculator_1: first calculator that computes a density descriptor, either
            a :py:class:`featomic.SphericalExpansion` or
            :py:class:`featomic.LodeSphericalExpansion`.
        :param calculator_2: optional second calculator that computes a density
            descriptor, either a :py:class:`featomic.SphericalExpansion` or
            :py:class:`featomic.LodeSphericalExpansion`. If ``None``, the equivariant
            power spectrum is computed as the auto-correlation of the first calculator.
            Defaults to ``None``.
        :param neighbor_types: List of ``"neighbor_type"`` to use in the properties of
            the output. This option might be useful when running the calculation on
            subset of a whole dataset and trying to join along the ``sample`` dimension
            after the calculation. If ``None``, blocks are filled with
            ``"neighbor_type"`` found in the systems. This parameter is only used if
            ``neighbors_to_properties=True`` is passed to the :py:meth:`compute` method.
        :param dtype: the scalar type to use to store coefficients
        :param device: the computational device to use for calculations.
        """

        super().__init__()
        self.calculator_1 = calculator_1
        self.calculator_2 = calculator_2
        self.neighbor_types = neighbor_types
        self.dtype = dtype
        self.device = device

        supported_calculators = ["lode_spherical_expansion", "spherical_expansion"]

        if self.calculator_1.c_name not in supported_calculators:
            raise ValueError(
                f"Only [{', '.join(supported_calculators)}] are supported for "
                f"`calculator_1`, got '{self.calculator_1.c_name}'"
            )

        parameters_1 = json.loads(calculator_1.parameters)

        if self.calculator_2 is None:
            parameters_2 = parameters_1
        else:
            if self.calculator_2.c_name not in supported_calculators:
                raise ValueError(
                    f"Only [{', '.join(supported_calculators)}] are supported for "
                    f"`calculator_2`, got '{self.calculator_2.c_name}'"
                )

            parameters_2 = json.loads(calculator_2.parameters)
            if parameters_1["basis"]["type"] != "TensorProduct":
                raise ValueError(
                    "only 'TensorProduct' basis is supported for calculator_1"
                )

            if parameters_2["basis"]["type"] != "TensorProduct":
                raise ValueError(
                    "only 'TensorProduct' basis is supported for calculator_2"
                )

        self._cg_product = ClebschGordanProduct(
            max_angular=parameters_1["basis"]["max_angular"]
            + parameters_2["basis"]["max_angular"],
            cg_backend=None,
            keys_filter=_filter_redundant_keys,
            arrays_backend=None,
            dtype=dtype,
            device=device,
        )

    @property
    def name(self):
        """Name of this calculator."""
        return "EquivariantPowerSpectrum"

    def compute(
        self,
        systems: Union[IntoSystem, List[IntoSystem]],
        selected_keys: Optional[Labels] = None,
        selected_samples: Optional[Labels] = None,
        neighbors_to_properties: bool = False,
    ) -> TensorMap:
        """
        Computes an equivariant power spectrum, also called "Lambda-SOAP" when doing a
        self-correlation of the SOAP density.

        First computes a :py:class:`SphericalExpansion` density descriptor of body order
        2.

        Before performing the Clebsch-Gordan tensor product, the spherical expansion
        density can be densified by moving the key dimension "neighbor_type" to the
        block properties. This is controlled by the ``neighbors_to_properties``
        parameter. Depending on the specific systems descriptors are being computed for,
        the sparsity or density of the density can affect the computational cost of the
        Clebsch-Gordan tensor product.

        If ``neighbors_to_properties=True`` and ``neighbor_types`` have been passed to
        the constructor, property dimensions are created for all of these global atom
        types when moving the key dimension to properties. This ensures that the output
        properties dimension is of consistent size across all systems passed in
        ``systems``.

        Finally a single Clebsch-Gordan tensor product is taken to produce a body order
        3 equivariant power spectrum.

        :param selected_keys: :py:class:`Labels`, the output keys to computed. If
            ``None``, all keys are computed. Subsets of key dimensions can be passed to
            compute output blocks that match in these dimensions.
        :param selected_samples: :py:class:`Labels`, Set of samples on which to run the
            calculation. Use ``None`` to run the calculation on all samples in
            the systems (this is the default). Gets passed to ``calculator_1``  and
            ``calculator_2``, therefore requiring that both calculators support sample
            selection.
        :param neighbors_to_properties: :py:class:`bool`, if true, densifies the
            spherical expansion by moving key dimension "neighbor_type" to properties
            prior to performing the Clebsch Gordan product step. Defaults to false.

        :return: :py:class:`TensorMap`, the output equivariant power spectrum.
        """
        return self._equivariant_power_spectrum(
            systems=systems,
            selected_keys=selected_keys,
            selected_samples=selected_samples,
            neighbors_to_properties=neighbors_to_properties,
            compute_metadata=False,
        )

    def forward(
        self,
        systems: Union[IntoSystem, List[IntoSystem]],
        selected_keys: Optional[Labels] = None,
        selected_samples: Optional[Labels] = None,
        neighbors_to_properties: bool = False,
    ) -> TensorMap:
        """
        Calls the :py:meth:`compute` method.

        This is intended for :py:class:`torch.nn.Module` compatibility, and should be
        ignored in pure Python mode.

        See :py:meth:`compute` for a full description of the parameters.
        """
        return self.compute(
            systems=systems,
            selected_keys=selected_keys,
            selected_samples=selected_samples,
            neighbors_to_properties=neighbors_to_properties,
        )

    def compute_metadata(
        self,
        systems: Union[IntoSystem, List[IntoSystem]],
        selected_keys: Optional[Labels] = None,
        selected_samples: Optional[Labels] = None,
        neighbors_to_properties: bool = False,
    ) -> TensorMap:
        """
        Returns the metadata-only :py:class:`TensorMap` that would be output by the
        function :py:meth:`compute` for the same calculator under the same settings,
        without performing the actual Clebsch-Gordan tensor products in the second step.

        See :py:meth:`compute` for a full description of the parameters.
        """
        return self._equivariant_power_spectrum(
            systems=systems,
            selected_keys=selected_keys,
            selected_samples=selected_samples,
            neighbors_to_properties=neighbors_to_properties,
            compute_metadata=True,
        )

    def _equivariant_power_spectrum(
        self,
        systems: Union[IntoSystem, List[IntoSystem]],
        selected_keys: Optional[Labels],
        selected_samples: Optional[Labels],
        neighbors_to_properties: bool,
        compute_metadata: bool,
    ) -> TensorMap:
        """
        Computes the equivariant power spectrum, either fully or just metadata
        """
        # Compute density
        density_1 = self.calculator_1.compute(
            systems, selected_samples=selected_samples
        )

        if self.calculator_2 is None:
            density_2 = density_1
        else:
            density_2 = self.calculator_2.compute(
                systems, selected_samples=selected_samples
            )

        # Rename "neighbor_type" dimension so they are correlated
        density_1 = operations.rename_dimension(
            density_1, "keys", "neighbor_type", "neighbor_1_type"
        )
        density_2 = operations.rename_dimension(
            density_2, "keys", "neighbor_type", "neighbor_2_type"
        )
        density_1 = operations.rename_dimension(density_1, "properties", "n", "n_1")
        density_2 = operations.rename_dimension(density_2, "properties", "n", "n_2")

        if neighbors_to_properties:
            if self.neighbor_types is None:  # just move neighbor type
                keys_to_move_1 = "neighbor_1_type"
                keys_to_move_2 = "neighbor_2_type"
            else:  # use the user-specified types
                values = _dispatch.list_to_array(
                    array=density_1.keys.values,
                    data=[[t] for t in self.neighbor_types],
                )
                keys_to_move_1 = Labels(names="neighbor_1_type", values=values)
                keys_to_move_2 = Labels(names="neighbor_2_type", values=values)

            density_1 = density_1.keys_to_properties(keys_to_move_1)
            density_2 = density_2.keys_to_properties(keys_to_move_2)

        # Compute the power spectrum
        if compute_metadata:
            pow_spec = self._cg_product.compute_metadata(
                tensor_1=density_1,
                tensor_2=density_2,
                o3_lambda_1_new_name="l_1",
                o3_lambda_2_new_name="l_2",
                selected_keys=selected_keys,
            )
        else:
            pow_spec = self._cg_product.compute(
                tensor_1=density_1,
                tensor_2=density_2,
                o3_lambda_1_new_name="l_1",
                o3_lambda_2_new_name="l_2",
                selected_keys=selected_keys,
            )

        # Move the CG combination info keys to properties
        pow_spec = pow_spec.keys_to_properties(["l_1", "l_2"])

        return pow_spec



"""
Private module containing helper functions for public module
:py:mod:`_correlate_density` that compute Clebsch-gordan tensor products of
:py:class:`TensorMap`.
"""

from typing import List, Tuple

from . import _coefficients, _dispatch
from ._backend import Array, Labels, TensorBlock, TensorMap


# ======================================== #
# ===== Functions to handle metadata ===== #
# ======================================== #


class Combination:
    """
    Small class to store together the index of two blocks to combine and the
    set of ``o3_lambdas`` created by this combination.
    """

    def __init__(self, first: int, second: int, o3_lambdas: List[int]):
        self.first = first
        self.second = second
        self.o3_lambdas = o3_lambdas


def _compute_output_keys(
    keys_1: Labels,
    keys_2: Labels,
    o3_lambda_1_new_name: str,
    o3_lambda_2_new_name: str,
) -> Tuple[Labels, List[Tuple[int, int]]]:
    """
    Computes the output keys from the full CG tensor product of ``keys_1`` and
    ``keys_2``.

    The output keys are the full product of the keys in ``keys_1`` and ``keys_2``, where
    Clebsch-Gordan combination rules are used to determine the new angular order and
    parity of output blocks. Key dimensions "o3_lambda" and "o3_sigma" must therefore be
    present in both input keys.

    ``o3_lambda_1_new_name`` and ``o3_lambda_2_new_name`` are the names of output key
    dimensions which store the "o3_lambda" values of the blocks in ``keys_1`` and
    ``keys_2`` respectively.

    For keys that are not named "o3_lambda" or "o3_sigma", we separate keys that exist
    in both inputs (matching dimensions) and dimensions that only exist in one of the
    inputs. Then, we only take correlations between blocks that have the same values for
    all matching keys, and perform a full tensor product for all non-matching
    dimensions.

    Returned is a :py:class:`Labels` object for the output keys, and a list of the pairs
    of indices in ``keys_1`` and ``keys_2`` that combine to form each key in the output
    :py:class:`Labels`.

    :param keys_1: :py:class:`Labels`, the keys of the first tensor.
    :param keys_2: :py:class:`Labels`, the keys of the second tensor.
    :param o3_lambda_1_new_name: str, the name of the output key dimension storing the
        "o3_lambda" values of the blocks in ``keys_1``.
    :param o3_lambda_2_new_name: str, the name of the output key dimension storing the
        "o3_lambda" values of the blocks in ``keys_2``.
    """

    # Define the names of the 'standard' (required) CG key names
    cg_names = ["o3_lambda", "o3_sigma"]
    for name in cg_names:
        assert name in keys_1.names, f"`keys_1` must contain dimension {name}"
        assert name in keys_2.names, f"`keys_2` must contain dimension {name}"

    # Define the names and column idxs of the 'match' keys, if any
    match_names: List[str] = []
    for name in keys_1.names:
        if name in cg_names:
            continue
        if name in keys_2.names:
            match_names.append(name)
    match_idxs_1: List[int] = [keys_1.names.index(name) for name in match_names]
    match_idxs_2: List[int] = [keys_2.names.index(name) for name in match_names]

    # keys_1: define the names and column idxs of the 'other' keys, if any
    other_names_1: List[str] = []
    for name in keys_1.names:
        if name in cg_names or name in match_names:
            continue
        other_names_1.append(name)
    other_idxs_1: List[int] = [keys_1.names.index(name) for name in other_names_1]

    # keys_2: define the names and column idxs of the 'other' keys, if any
    other_names_2: List[str] = []
    for name in keys_2.names:
        if name in cg_names or name in match_names:
            continue
        other_names_2.append(name)
    other_idxs_2: List[int] = [keys_2.names.index(name) for name in other_names_2]

    # Build the new names: this enforces a certain order in the output keys
    output_names: List[str] = (
        cg_names
        + match_names
        + other_names_1
        + [o3_lambda_1_new_name]
        + other_names_2
        + [o3_lambda_2_new_name]
    )

    # Build the new keys by combination
    output_key_values: List[List[int]] = []
    combinations: List[Tuple[int, int]] = []
    for key_1_i in range(len(keys_1)):
        for key_2_i in range(len(keys_2)):
            # Get the keys
            key_1 = keys_1.entry(key_1_i)
            key_2 = keys_2.entry(key_2_i)

            # If they don't match in the match dimensions, skip combination
            if not _dispatch.all(
                key_1.values[match_idxs_1] == key_2.values[match_idxs_2]
            ):
                continue

            # Unpack standard key values
            o3_lambda_1, o3_sigma_1 = int(key_1["o3_lambda"]), int(key_1["o3_sigma"])
            o3_lambda_2, o3_sigma_2 = int(key_2["o3_lambda"]), int(key_2["o3_sigma"])

            # Create all non-zero angular channel combinations
            for o3_lambda in range(
                abs(o3_lambda_1 - o3_lambda_2), abs(o3_lambda_1 + o3_lambda_2) + 1
            ):
                # Calculate new sigma
                o3_sigma = int(
                    o3_sigma_1
                    * o3_sigma_2
                    * (-1) ** (o3_lambda_1 + o3_lambda_2 + o3_lambda)
                )

                # Build the new keys values
                new_vals: List[int] = (
                    [o3_lambda, o3_sigma]  # standard CG dims
                    + _dispatch.to_int_list(key_1.values[match_idxs_1])  # match dims
                    + _dispatch.to_int_list(key_1.values[other_idxs_1])  # other dims 1
                    + [o3_lambda_1]  # lambda from block_1
                    + _dispatch.to_int_list(key_2.values[other_idxs_2])  # other dims 2
                    + [o3_lambda_2]  # lambda from block_2
                )
                output_key_values.append(new_vals)
                combinations.append((key_1_i, key_2_i))

    # Define new keys as the full product of keys_1 and keys_2
    output_keys = Labels(
        names=output_names,
        values=_dispatch.int_array_like(output_key_values, like=keys_1.values),
    )

    return output_keys, combinations


def _group_combinations_of_same_blocks(
    output_keys: Labels,
    combinations: List[Tuple[int, int]],
) -> Tuple[Labels, List[Combination]]:
    """
    Re-order output keys in such a way that we have all the different ``o3_lambda``
    created by the same pair of block consecutive in ``output_keys``.

    This function simultaneously returns the new ``output_keys``, and a list of
    combinations in the same order as the keys. Iterating over the returned combinations
    and then over the ``o3_lambdas`` for each combination will produce blocks in the
    same order as the output keys.
    """
    # (block_1, block_2) => list of index in the output keys
    # this is emulating a dict by separating keys & values since TorchScript does not
    # support Tuple[int, int] as Dict keys. `groups` is the dict keys and `groups_keys`
    # the dict values.
    groups: List[Tuple[int, int]] = []
    groups_keys: List[List[int]] = []
    for key_i, blocks in enumerate(combinations):
        group_id = -1
        for existing_id, existing in enumerate(groups):
            if existing == blocks:
                group_id = existing_id
                break

        if group_id == -1:
            group_id = len(groups)
            groups.append(blocks)
            keys_idx: List[int] = []
            groups_keys.append(keys_idx)

        groups_keys[group_id].append(key_i)

    all_o3_lambdas = output_keys.column("o3_lambda")

    keys_values: List[List[int]] = []
    combinations: List[Combination] = []
    for (block_1, block_2), keys_idx in zip(groups, groups_keys, strict=True):
        o3_lambdas: List[int] = []
        for key_i in keys_idx:
            keys_values.append(_dispatch.to_int_list(output_keys.values[key_i]))
            o3_lambdas.append(int(all_o3_lambdas[key_i]))

        combinations.append(Combination(block_1, block_2, o3_lambdas))

    output_keys = Labels(
        output_keys.names,
        _dispatch.int_array_like(keys_values, like=output_keys.values),
    )

    return output_keys, combinations


# ================================================================ #
# ======== Functions to take CG tensor products of blocks ======== #
# ================================================================ #


def cg_tensor_product_blocks(
    block_1: TensorBlock,
    block_2: TensorBlock,
    o3_lambdas: List[int],
    cg_coefficients: TensorMap,
    cg_backend: str,
) -> List[TensorBlock]:
    """
    For a given pair of TensorBlocks and desired angular channels, combines the values
    arrays and returns a new TensorBlock.

    If the samples of the two blocks are different, we assume that one of the sample
    names are a subset of the other sample names. We then match sample values for
    this subset, and reshape the block with the smallest sample names to have the
    same number of samples as the other one.
    """
    # Expand the blocks along the samples axis by matching samples dimensions.
    # TODO: add a new operation in metatensor to do this
    block_1, block_2 = _match_samples_of_blocks(block_1, block_2)

    # Compute the new properties dimensions for the output TensorBlocks
    properties = _compute_labels_full_cartesian_product(
        block_1.properties, block_2.properties
    )

    # Do the CG combination - single center so no shape pre-processing required
    combined_values = _coefficients.cg_tensor_product(
        block_1.values, block_2.values, o3_lambdas, cg_coefficients, cg_backend
    )

    # Create the TensorBlocks
    results: List[TensorBlock] = []
    for values, o3_lambda in zip(combined_values, o3_lambdas, strict=True):
        block = TensorBlock(
            values=values,
            samples=block_1.samples,
            components=[
                Labels(
                    names=["o3_mu"],
                    values=_dispatch.int_range_like(
                        min_val=-o3_lambda,
                        max_val=o3_lambda + 1,
                        like=block_1.samples.values,
                    ).reshape(-1, 1),
                ),
            ],
            properties=properties,
        )
        results.append(block)

    return results


def _match_samples_of_blocks(
    block_1: TensorBlock, block_2: TensorBlock
) -> Tuple[TensorBlock, TensorBlock]:
    """
    Extend the samples axes of either ``block_1`` or ``block_2`` to match the samples of
    the other.

    Assumes that the samples dimensions of the block with fewer dimensions are a subset
    of the dimensions of the other. If the dimensions are not a subset, an error is
    raised.
    """
    # The number of dimensions are the same
    if len(block_1.samples.names) == len(block_2.samples.names):
        if block_1.samples == block_2.samples:  # nothing needs to be done
            return block_1, block_2

        # Find the union of samples and broadcast both blocks along samples axis
        new_blocks: List[TensorBlock] = []
        union, mapping_1, mapping_2 = block_1.samples.union_and_mapping(block_2.samples)
        for block, mapping in [(block_1, mapping_1), (block_2, mapping_2)]:
            new_block_vals = _dispatch.zeros_like(
                block.values, [len(union)] + [i for i in block.values.shape[1:]]
            )
            new_block_vals[mapping] = block.values
            new_block = TensorBlock(
                values=new_block_vals,
                samples=union,
                components=block.components,
                properties=block.properties,
            )
            new_blocks.append(new_block)

        block_1, block_2 = new_blocks

        return block_1, block_2

    # Otherwise, the samples dimensions of one block is assumed (and checked) to be a
    # subset of the other.
    assert len(block_1.samples.names) != len(block_2.samples.names)

    # First find the block with fewer dimensions. Reorder to have this block on the
    # 'left' for simplicity, but record the original order for the final output
    swapped = False
    if len(block_1.samples.names) > len(block_2.samples.names):
        block_1, block_2 = block_2, block_1
        swapped = True

    for name in block_1.samples.names:
        if name not in block_2.samples.names:
            raise ValueError(
                "Samples dimensions of the two blocks are not compatible."
                " The block with fewer samples dimensions must be a subset"
                f" of the other block's dimensions. Got: {block_1.samples.names}"
                f" and {block_2.samples.names}. Offending dimension: {name}"
            )

    # Broadcast the values of block_1 along the samples dimensions to match those of
    # block_2
    dims_2 = [block_2.samples.names.index(name) for name in block_1.samples.names]
    matches: List[Array] = _dispatch.where(
        _dispatch.all(
            block_2.samples.values[:, dims_2][:, None] == block_1.samples.values,
            axis=2,
        )
    )

    # Build new block_1
    block_1 = TensorBlock(
        values=block_1.values[matches[1]],
        samples=Labels(block_2.samples.names, block_2.samples.values[matches[0]]),
        components=block_1.components,
        properties=block_1.properties,
    )

    # Check if we need to slice block_2 samples due to samples in block_2 not being
    # matched in block_1
    if (
        len(matches[0]) == len(block_2.samples)
        and not _dispatch.all(
            matches[0]
            == _dispatch.int_range_like(
                0, len(block_2.samples), like=block_2.samples.values
            )
        )
    ) or len(matches[0]) != len(block_2.samples):
        # Build new block_2
        block_2 = TensorBlock(
            values=block_2.values[matches[0]],
            samples=Labels(block_2.samples.names, block_2.samples.values[matches[0]]),
            components=block_2.components,
            properties=block_2.properties,
        )

    if swapped:
        return block_2, block_1

    return block_1, block_2


def _compute_labels_full_cartesian_product(
    labels_1: Labels,
    labels_2: Labels,
) -> Labels:
    """
    Computes the full cartesian product of two arbitrary :py:class:`Labels` objects.

    In contrast to the :py:func:`_precompute_keys_full_product` function, this function
    has no awareness of angular channel combination rules or value matching. It simply
    takes all combinations of the labels entries in ``labels_1`` and ``labels_2``.

    This function assumes that there are no shared labels dimension names between
    ``labels_1`` and ``labels_2``.
    """
    # Check for no shared labels dimensions
    for name in labels_1.names:
        assert name not in labels_2.names, (
            "`labels_1` and `labels_2` must not have a dimension ({name}) in common"
        )
    # Create the new labels names by concatenating the names of the two input labels
    labels_names: List[str] = labels_1.names + labels_2.names

    # create cross product list of indices in a torch-scriptable way of
    # [[i, j] for i in range(len(labels_1.values)) for j in
    #             range(len(labels_2.values))]
    # [0, 1, 2], [0, 1] -> [[0, 1], [0, 2], [1, 0], [1, 1], [2, 0], [2, 1]]

    labels_1_labels_2_product_idx = _dispatch.cartesian_prod(
        _dispatch.int_range_like(0, len(labels_1.values), like=labels_1.values),
        _dispatch.int_range_like(0, len(labels_2.values), like=labels_2.values),
    )
    return Labels(
        names=labels_names,
        values=_dispatch.int_array_like(
            [
                _dispatch.to_int_list(labels_1.values[indices[0]])
                + _dispatch.to_int_list(labels_2.values[indices[1]])
                for indices in labels_1_labels_2_product_idx
            ],
            like=labels_1.values,
        ),
    )



import abc
from typing import Dict, List, Optional

import numpy as np


try:
    import scipy.integrate
    import scipy.optimize
    import scipy.special

    HAS_SCIPY = True
except ImportError:
    HAS_SCIPY = False


class RadialBasis(metaclass=abc.ABCMeta):
    """
    Base class representing a set of radial basis functions, indexed by a radial index
    ``n``.

    You can inherit from this class to define new custom radial basis function, by
    implementing the :py:meth:`compute_primitive` method. If needed,
    :py:meth:`finite_differences_derivative` can be used to compute the derivatives of a
    radial basis.

    Overriding :py:attr:`integration_radius` can be useful to control the integration
    radius when evaluating the radial integral numerically. See this :ref:`explanation
    <radial-integral>` for more information.

    If the new radial basis function has corresponding hyper parameters in the native
    calculators, you should also implement :py:meth:`get_hypers`.
    """

    def __init__(self, *, max_radial: int, radius: float):
        """
        :parameter max_radial: maximal radial basis index to include (there will be
            ``N = max_radial + 1`` basis functions overall)
        :parameter radius: radius of the radial basis. For local spherical expansions,
            this is typically the same as the spherical cutoff radius.
        """
        self.max_radial = int(max_radial)
        self.radius = float(radius)

        assert self.max_radial >= 0
        assert self.radius > 0.0
        self._orthonormalization_matrix = None

    def _featomic_hypers(self):
        return self.get_hypers()

    def get_hypers(self):
        """
        Return the native hyper parameters corresponding to this set of basis functions
        """
        raise NotImplementedError(
            f"This radial basis function ({self.__class__.__name__}) does not have "
            "matching hyper parameters in the native calculators. It should be used "
            "through one of the spliner class instead of directly."
        )

    @property
    def size(self) -> int:
        """Get the size of the basis set (i.e. the total number of basis functions)"""
        return self.max_radial + 1

    @property
    def integration_radius(self) -> float:
        """
        Get the radius to use for numerical evaluation of radial integrals.

        This default to the ``radius`` given as a class parameter, but can be overridden
        by child classes as needed.
        """
        return self.radius

    @abc.abstractmethod
    def compute_primitive(
        self, positions: np.ndarray, n: int, *, derivative: bool
    ) -> np.ndarray:
        """
        Evaluate the primitive (not normalized not orthogonalized) radial basis for
        index ``n`` on grid points at the given ``positions``.

        :param n: index of the radial basis to evaluate
        :param positions: positions of the grid points where the basis should be
            evaluated
        :param derivative: should this function return the values of the radial basis or
            its derivatives.
        :return: the values (or derivative) of radial basis on the grid points
        """

    def compute(self, positions: np.ndarray, *, derivative: bool) -> np.ndarray:
        """
        Evaluate the orthogonalized and normalized radial basis on grid points at the
        given ``positions``. The returned array contains all the radial basis, from
        ``n=0`` to ``n = max_radial``.

        :param positions: positions of the grid points where the basis should be
            evaluated
        :param derivative: should this function return the values of the radial basis or
            its derivatives.
        :return: the values (or derivative) of radial basis on the grid points
        """
        if self._orthonormalization_matrix is None:
            self._orthonormalization_matrix = self._get_orthonormalization_matrix()

        basis = np.vstack(
            [
                self.compute_primitive(positions, n, derivative=derivative)
                for n in range(self.size)
            ]
        )
        return (self._orthonormalization_matrix @ basis).T

    def finite_differences_derivative(
        self,
        positions: np.ndarray,
        n: int,
        *,
        displacement=1e-6,
    ) -> np.ndarray:
        """
        Helper function to compute derivate of the radial function using finite
        differences. This can be used by child classes to implement the
        ``derivative=True`` branch of :py:meth:`compute` function.
        """
        value_pos = self.compute_primitive(
            positions + displacement / 2, n, derivative=False
        )
        value_neg = self.compute_primitive(
            positions - displacement / 2, n, derivative=False
        )

        return (value_pos - value_neg) / displacement

    def _get_orthonormalization_matrix(self) -> np.ndarray:
        """
        Compute the ``(self.size, self.size)`` orthonormalization matrix for this radial
        basis using numerical integration.
        """
        if not HAS_SCIPY:
            raise ValueError("Orthonormalization requires scipy!")

        gram_matrix = self._get_gram_matrix()

        # Get the normalization constants from the diagonal entries
        normalizations = np.zeros(self.size)

        for n in range(self.size):
            normalizations[n] = 1 / np.sqrt(gram_matrix[n, n])
            # Rescale gram matrix to be defined in terms of the normalized
            # (but not yet orthonormalized) basis functions
            gram_matrix[n, :] *= normalizations[n]
            gram_matrix[:, n] *= normalizations[n]

        eigvals, eigvecs = np.linalg.eigh(gram_matrix)
        if np.any(eigvals < 1e-12):
            raise ValueError(
                "Unable to orthonormalize the radial basis, gram matrix is singular. "
                "You can try decreasing the number of radial basis function, or "
                "changing some of the basis function parameters"
            )

        orthonormalization_matrix = (
            eigvecs @ np.diag(np.sqrt(1.0 / eigvals)) @ eigvecs.T
        )

        # Rescale the orthonormalization matrix so that it
        # works with respect to the primitive (i.e. not normalized)
        # radial basis functions
        for n in range(self.size):
            orthonormalization_matrix[:, n] *= normalizations[n]

        return orthonormalization_matrix

    def _get_gram_matrix(self) -> np.ndarray:
        """compute the Gram matrix of the current basis."""
        # Gram matrix (also called overlap matrix or inner product matrix)
        gram_matrix = np.zeros((self.size, self.size))

        def integrand(
            positions: np.ndarray,
            n1: int,
            n2: int,
        ) -> np.ndarray:
            r1 = self.compute_primitive(positions, n1, derivative=False)
            r2 = self.compute_primitive(positions, n2, derivative=False)
            return positions**2 * r1 * r2

        for n1 in range(self.size):
            for n2 in range(self.size):
                gram_matrix[n1, n2] = scipy.integrate.quad(
                    func=integrand,
                    a=0,
                    b=self.integration_radius,
                    args=(n1, n2),
                )[0]

        return gram_matrix


class Gto(RadialBasis):
    r"""
    Gaussian Type Orbital (GTO) radial basis.

    It is defined as

    .. math::

        R_n(r) = r^n e^{-\frac{r^2}{2\sigma_n^2}},

    where :math:`\sigma_n = \sqrt{n} r_0 / N` with :math:`r_0` being the basis
    ``radius`` and :math:`N` the number of radial basis functions.
    """

    def __init__(self, *, max_radial: int, radius: Optional[float] = None):
        """
        :parameter max_radial: maximal radial basis index to include (there will be
            ``N = max_radial + 1`` basis functions overall)
        :parameter radius: radius of the GTO basis functions. This is only required for
            LODE spherical expansion or splining the radial integral.
        """
        if radius is None:
            super().__init__(max_radial=max_radial, radius=float("inf"))
        else:
            super().__init__(max_radial=max_radial, radius=radius)

        self._gto_sigmas = np.ones(self.size, dtype=np.float64)

        if radius is not None:
            self._gto_radius = float(radius)
            for n in range(1, self.size):
                self._gto_sigmas[n] = np.sqrt(n)
            self._gto_sigmas *= self._gto_radius / self.size
        else:
            self._gto_radius = None

    def get_hypers(self):
        hypers = {
            "type": "Gto",
            "max_radial": self.max_radial,
        }

        if self._gto_radius is not None:
            hypers["radius"] = self._gto_radius

        return hypers

    @property
    def integration_radius(self) -> float:
        if self._gto_radius is None:
            raise ValueError(
                "`radius` needs to be specified for numerical evaluation "
                "of the GTO radial basis"
            )

        # We would ideally infinity as the ``integration_radius``, but this leads to
        # problems when calculating the radial integral with `quad`. So we pick
        # something large enough instead
        return 5 * self._gto_radius

    def compute_primitive(
        self, positions: np.ndarray, n: int, *, derivative: bool
    ) -> np.ndarray:
        if self._gto_radius is None:
            raise ValueError(
                "`radius` needs to be specified for numerical evaluation "
                "of the GTO radial basis"
            )

        values = positions**n * np.exp(-0.5 * (positions / self._gto_sigmas[n]) ** 2)
        if derivative:
            return (
                n / positions * values - positions / self._gto_sigmas[n] ** 2 * values
            )

        return values


class Monomials(RadialBasis):
    r"""
    Monomial radial basis, consisting of functions:

    .. math::
        R_{nl}(r) = r^{l+2n}

    These capture precisely the radial dependence if we compute the Taylor expansion of
    a generic function defined in 3D space.

    :parameter angular_channel: index of the angular channel associated with this radial
        basis, i.e. :math:`l` in the equation above.
    """

    def __init__(self, *, angular_channel: int, max_radial: int, radius: float):
        super().__init__(max_radial=max_radial, radius=radius)
        self.angular_channel = int(angular_channel)
        assert self.angular_channel >= 0

    def compute_primitive(
        self, positions: np.ndarray, n: int, *, derivative: bool
    ) -> np.ndarray:
        ell = self.angular_channel
        if derivative:
            return (ell + 2 * n) * positions ** (ell + 2 * n - 1)
        else:
            return positions ** (ell + 2 * n)


class SphericalBessel(RadialBasis):
    r"""Spherical Bessel functions as a radial basis.

    This is used among others in the `Laplacian eigenstate
    <https://doi.org/10.1063/5.0124363>`_ basis. The basis functions have the following
    form:

    .. math::

        R_{ln}(r) = j_l \left( \frac{r}{r_0} \text{zero}(J_l, n) \right)

    where :math:`j_l` is the spherical bessel function of the first kind of order
    :math:`l`, :math:`r_0` is the basis function ``radius``, and :math:`\text{zero}(J_l,
    n)` is the :math:`n`-th zero of (non-spherical) bessel function of first kind and
    order :math:`l`.

    :parameter angular_channel: index of the angular channel associated with this radial
        basis, i.e. :math:`l` in the equation above.
    """

    def __init__(self, *, angular_channel: int, max_radial: int, radius: float):
        super().__init__(max_radial=max_radial, radius=radius)
        self.angular_channel = int(angular_channel)
        assert self.angular_channel >= 0

        # this is computing all roots for all `l` up to `angular_channel` to then throw
        # away most of them. Maybe there is a better way to do this
        self._roots = SphericalBessel._compute_zeros(angular_channel + 1, self.size)

    def compute_primitive(
        self, positions: np.ndarray, n: int, *, derivative: bool
    ) -> np.ndarray:
        ell = self.angular_channel
        values = scipy.special.spherical_jn(
            ell, positions * self._roots[ell, n] / self.radius, derivative=derivative
        )
        if derivative:
            values *= self._roots[ell, n] / self.radius

        return values

    @staticmethod
    def _compute_zeros(angular_size: int, radial_size: int) -> np.ndarray:
        """Zeros of spherical bessel functions.

        Code is taken from the `Scipy Cookbook <spc>`_

        .. _spc: https://scipy-cookbook.readthedocs.io/items/SphericalBesselZeros.html

        :parameter angular_size: number of angular components
        :parameter radial_size: number of radial components
        :returns: computed zeros of the spherical bessel functions
        """

        def Jn(r: float, ell: int) -> float:
            return np.sqrt(np.pi / (2 * r)) * scipy.special.jv(ell + 0.5, r)

        def Jn_zeros(angular_size: int, radial_size: int) -> np.ndarray:
            zeros_j = np.zeros((angular_size, radial_size), dtype=np.float64)
            zeros_j[0] = np.arange(1, radial_size + 1) * np.pi
            points = np.arange(1, radial_size + angular_size + 1) * np.pi
            roots = np.zeros(radial_size + angular_size, dtype=np.float64)
            for ell in range(1, angular_size):
                for j in range(radial_size + angular_size - ell):
                    roots[j] = scipy.optimize.brentq(
                        Jn, points[j], points[j + 1], (ell,)
                    )
                points = roots
                zeros_j[ell][:radial_size] = roots[:radial_size]
            return zeros_j

        return Jn_zeros(angular_size, radial_size)


########################################################################################
########################################################################################


class ExpansionBasis(metaclass=abc.ABCMeta):
    """
    Base class representing a set of basis functions used by spherical expansions.

    A full basis typically uses both a set of radial basis functions, and angular basis
    functions; combined in various ways. The angular basis functions are almost always
    spherical harmonics, while the radial basis function can be freely picked.

    You can inherit from this class to define new sets of basis functions, implementing
    :py:meth:`get_hypers` to create the right hyper parameters for the underlying native
    calculator, as well as :py:meth:`angular_channels` and :py:meth:`radial_basis` to
    define the set of basis functions to use.
    """

    def _featomic_hypers(self):
        return self.get_hypers()

    def get_hypers(self):
        """
        Return the native hyper parameters corresponding to this set of basis functions
        """
        raise NotImplementedError(
            f"This basis functions set ({self.__class__.__name__}) does not have "
            "matching hyper parameters in the native calculators. It should be used "
            "through one of the spliner class instead of directly."
        )

    @abc.abstractmethod
    def angular_channels(self) -> List[int]:
        """Get the list of angular channels that are included in the expansion basis."""

    @abc.abstractmethod
    def radial_basis(self, angular: int) -> RadialBasis:
        """Get the radial basis used for a given ``angular`` channel"""


class TensorProduct(ExpansionBasis):
    r"""
    Basis function set combining spherical harmonics with a radial basis functions set,
    taking all possible combinations of radial and angular basis function.

    Using ``N`` radial basis functions and ``L`` angular basis functions, this will
    create ``N x L`` basis functions (:math:`B_{nlm}`) for the overall expansion:

    .. math::

        B_{nlm}(\boldsymbol{r}) = R_{nl}(r)Y_{lm}(\hat{r}) \,
    """

    def __init__(
        self,
        *,
        max_angular: int,
        radial: RadialBasis,
        spline_accuracy: Optional[float] = 1e-8,
    ):
        """
        :param max_angular: Largest angular channel to include in the basis
        :param radial: radial basis to use for all angular channels
        :param spline_accuracy: requested accuracy of the splined radial integrals,
            defaults to 1e-8
        """
        self.max_angular = int(max_angular)
        self.radial = radial

        if spline_accuracy is None:
            self.spline_accuracy = None
        else:
            self.spline_accuracy = float(spline_accuracy)
            assert self.spline_accuracy > 0

        assert self.max_angular >= 0
        assert isinstance(self.radial, RadialBasis)

    def get_hypers(self):
        return {
            "type": "TensorProduct",
            "max_angular": self.max_angular,
            "radial": self.radial,
            "spline_accuracy": self.spline_accuracy,
        }

    def angular_channels(self) -> List[int]:
        return list(range(self.max_angular + 1))

    def radial_basis(self, angular: int) -> RadialBasis:
        return self.radial


class Explicit(ExpansionBasis):
    """
    An expansion basis where combinations of radial and angular functions is picked
    explicitly.

    The angular basis functions are still spherical harmonics, but only the degrees
    included as keys in ``by_angular`` will be part of the output. Each of these angular
    basis function can then be associated with a set of different radial basis function,
    potentially of different sizes.
    """

    def __init__(
        self,
        *,
        by_angular: Dict[int, RadialBasis],
        spline_accuracy: Optional[float] = 1e-8,
    ):
        """
        :param by_angular: definition of the radial basis for each angular channel to
            include.
        :param spline_accuracy: requested accuracy of the splined radial integrals,
            defaults to 1e-8
        """
        self.by_angular = by_angular

        if spline_accuracy is None:
            self.spline_accuracy = None
        else:
            self.spline_accuracy = float(spline_accuracy)
            assert self.spline_accuracy > 0

        for angular, radial in self.by_angular.items():
            assert angular >= 0
            assert isinstance(radial, RadialBasis)

    def get_hypers(self):
        return {
            "type": "Explicit",
            "by_angular": self.by_angular,
            "spline_accuracy": self.spline_accuracy,
        }

    def angular_channels(self) -> List[int]:
        return list(self.by_angular.keys())

    def radial_basis(self, angular: int) -> RadialBasis:
        return self.by_angular[angular]


class LaplacianEigenstate(ExpansionBasis):
    """
    The Laplacian eigenstate basis, introduced in https://doi.org/10.1063/5.0124363, is
    a set of basis functions for spherical expansion which is both *smooth* (in the same
    sense as the smoothness of a low-pass-truncated Fourier expansion) and *ragged*,
    using a different number of radial function for each angular channel. This is
    intended to obtain a more balanced smoothness level in the radial and angular
    direction for a given total number of basis functions.

    This expansion basis is not directly implemented in the native calculators, but is
    intended to be used with the :py:class:`featomic.splines.SoapSpliner` to create
    splines of the radial integrals.
    """

    def __init__(
        self,
        *,
        radius: float,
        max_radial: int,
        max_angular: Optional[int] = None,
        spline_accuracy: Optional[float] = 1e-8,
    ):
        """
        :param radius: radius of the basis functions
        :param max_radial: number of radial basis function for the ``L=0`` angular
            channel. All other angular channels will have fewer radial basis functions.
        :param max_angular: Truncate the set of radial functions at this angular
            channel. If ``None``, this will be set to a high enough value to include all
            basis functions with an Laplacian eigenvalue below the one for ``l=0,
            n=max_radial``.
        :param spline_accuracy: requested accuracy of the splined radial integrals,
            defaults to 1e-8
        """
        self.radius = float(radius)
        assert self.radius >= 0

        self.max_radial = int(max_radial)
        assert self.max_radial >= 0

        if max_angular is None:
            self.max_angular = self.max_radial
        else:
            self.max_angular = int(max_angular)
            assert self.max_angular >= 0

        if spline_accuracy is None:
            self.spline_accuracy = None
        else:
            self.spline_accuracy = float(spline_accuracy)
            assert self.spline_accuracy > 0

        # compute the zeros of the spherical Bessel functions
        zeros_ln = SphericalBessel._compute_zeros(
            self.max_angular + 1, self.max_radial + 1
        )

        # determine the eigenvalue cutoff
        eigenvalues = zeros_ln**2 / self.radius**2
        max_eigenvalue = eigenvalues[0, max_radial]

        # find the actual `max_angular` if the user did not specify one by repeatedly
        # increasing the size of `eigenvalues`, until we find an angular channel where
        # all eigenvalues are above the cutoff.
        if max_angular is None:
            while eigenvalues[-1, 0] < max_eigenvalue:
                self.max_angular += self.max_radial
                zeros_ln = SphericalBessel._compute_zeros(
                    self.max_angular + 1, self.max_radial + 1
                )
                eigenvalues = zeros_ln**2 / self.radius**2

            self.max_angular = len(np.where(eigenvalues[:, 0] <= max_eigenvalue)[0]) - 1
            assert self.max_angular >= 0

        by_angular = {}
        for angular in range(self.max_angular + 1):
            max_radial = len(np.where(eigenvalues[angular, :] <= max_eigenvalue)[0]) - 1
            by_angular[angular] = SphericalBessel(
                angular_channel=angular,
                max_radial=max_radial,
                radius=self.radius,
            )

        self._explicit = Explicit(
            by_angular=by_angular,
            spline_accuracy=self.spline_accuracy,
        )

    def get_hypers(self):
        return self._explicit.get_hypers()

    def angular_channels(self) -> List[int]:
        return self._explicit.angular_channels()

    def radial_basis(self, angular: int) -> RadialBasis:
        return self._explicit.radial_basis(angular)



import abc
import warnings
from typing import Optional

import numpy as np


try:
    import scipy.special

    HAS_SCIPY = True
except ImportError:
    HAS_SCIPY = False


class RadialScaling(metaclass=abc.ABCMeta):
    """
    Base class representing radial scaling of atomic densities.

    You can inherit from this class to define new custom radial scaling, implementing
    :py:meth:`compute` accordingly. If the new radial scaling has corresponding hyper
    parameters in the native calculators, you should also implement
    :py:meth:`get_hypers`.
    """

    def _featomic_hypers(self):
        return self.get_hypers()

    def get_hypers(self):
        """
        Return the native hyper parameters corresponding to this atomic density scaling
        """
        raise NotImplementedError(
            f"this density scaling ({self.__class__.__name__}) does not have matching "
            "hyper parameters in the native calculators"
        )

    @abc.abstractmethod
    def compute(self, positions: np.ndarray, *, derivative: bool) -> np.ndarray:
        """
        Compute the scaling function (or it's derivative) on grid points at the given
        ``positions``.

        :param positions: positions of grid point where to evaluate the radial scaling
        :param derivative: should this function return the values or the derivatives of
            the radial scaling?
        :returns: evaluated radial scaling function
        """


class Willatt2018(RadialScaling):
    r"""
    Radial density scaling as proposed in https://doi.org/10.1039/C8CP05921G by Willatt
    et. al.

    .. math::

        \text{scaling}(r) = \frac{c}{c + \left(\frac{r}{r_0}\right) ^ m}
    """

    def __init__(self, *, exponent: int, rate: float, scale: float):
        """
        :param exponent: :math:`m` in the formula above
        :param rate: :math:`c` in the formula above
        :param scale: :math:`r_0` in the formula above
        """
        self.exponent = int(exponent)
        self.rate = float(rate)
        self.scale = float(scale)

        assert self.exponent >= 0
        assert self.scale > 0
        assert self.rate >= 0

    def get_hypers(self):
        return {
            "type": "Willatt2018",
            "exponent": self.exponent,
            "rate": self.rate,
            "scale": self.scale,
        }

    def compute(self, positions: np.ndarray, *, derivative: bool) -> np.ndarray:
        if self.rate == 0:
            result = (self.scale / positions) ** self.exponent
            if derivative:
                result *= -self.exponent / positions
            return result
        elif self.exponent == 0:
            if derivative:
                return np.zeros_like(positions)
            else:
                return np.ones_like(positions)
        else:
            if derivative:
                self.rate / (self.rate + (positions / self.scale) ** self.exponent)
            else:
                rs = positions / self.scale
                denominator = (self.rate + rs**self.exponent) ** 2

                factor = -self.rate * self.exponent / self.scale

                return factor * rs ** (self.exponent - 1) / denominator


class AtomicDensity(metaclass=abc.ABCMeta):
    r"""
    Base class representing atomic densities.

    You can inherit from this class to define new custom densities, implementing
    :py:meth:`compute` accordingly. If the new density has corresponding hyper
    parameters in the native calculators, you should also implement
    :py:meth:`get_hypers`.

    All atomic densities are assumed to be invariant under rotation, and as such only
    defined as a function of the distance to the origin.

    The overall density around a central atom is the sum of the central atom's neighbors
    density, with some optional radial scaling.

    .. math::

        \rho_i(r) = \sum_j \text{scaling}(r_{ij}) \; g_j(r - r_{ij})

    where :math:`\text{scaling}(r)` is the scaling function, :math:`g_j` the density
    coming from neighbor :math:`j`, and :math:`r_{ij}` the distance between the center
    :math:`i` and neighbor :math:`j`.
    """

    def __init__(
        self,
        *,
        center_atom_weight: float = 1.0,
        scaling: Optional[RadialScaling] = None,
    ):
        r"""
        :param center_atom_weight: in density expansion, the central atom sees its own
            density, and is in this sense its own neighbor. Setting this weight to ``0``
            allows to disable this behavior and only expand the density of actual
            neighbors.
        :param scaling: optional radial scaling function. If this is left to ``None``,
            no radial scaling is applied.
        """
        self.center_atom_weight = float(center_atom_weight)
        self.scaling = scaling

        assert isinstance(self.scaling, (type(None), RadialScaling))

    @abc.abstractmethod
    def compute(self, positions: np.ndarray, *, derivative: bool) -> np.ndarray:
        """
        Compute the density (or it's derivative) around a single atom.

        The atom is located at ``position=0`` and the density is computed on multiple
        grid points at the given ``positions``. The computed density does not include
        any radial scaling or central atom weighting.

        :param positions: positions of grid point where to evaluate the atomic density
        :param derivative: should this function return the values or the derivatives of
            the density?
        :returns: evaluated atomic density on the grid
        """

    def get_hypers(self):
        """
        Return the native hyper parameters corresponding to this atomic density
        """
        raise NotImplementedError(
            f"This density ({self.__class__.__name__}) does not have matching "
            "hyper parameters in the native calculators. It should be used "
            "through one of the spliner class instead of directly."
        )

    def _featomic_hypers(self):
        """
        Return the native hyper parameters corresponding to this atomic density.
        """
        return {
            **self.get_hypers(),
            "center_atom_weight": self.center_atom_weight,
            "scaling": self.scaling,
        }


class DiracDelta(AtomicDensity):
    r"""Delta atomic densities of the form :math:`g(r)=\delta(r)`."""

    def __init__(
        self,
        *,
        center_atom_weight: float = 1.0,
        scaling: Optional[RadialScaling] = None,
    ):
        super().__init__(center_atom_weight=center_atom_weight, scaling=scaling)

    def get_hypers(self):
        return {"type": "DiracDelta"}

    def compute(self, positions: np.ndarray, *, derivative: bool) -> np.ndarray:
        if derivative:
            return np.zeros_like(positions)
        else:
            result = np.zeros_like(positions)
            result[result == 0.0] = 1.0
            return result


class Gaussian(AtomicDensity):
    r"""Gaussian atomic density function.

    In featomic, we use the convention

    .. math::

        g(r) = \frac{1}{(\pi \sigma^2)^{3/4}}e^{-\frac{r^2}{2\sigma^2}} \,.

    The prefactor was chosen such that the "L2-norm" of the Gaussian

    .. math::

            \|g\|^2 = \int \mathrm{d}^3\boldsymbol{r} |g(r)|^2 = 1\,,

    The derivatives of the Gaussian atomic density with respect to the position is

    .. math::

        g^\prime(r) =
            \frac{\partial g(r)}{\partial r} = \frac{-r}{\sigma^2(\pi
            \sigma^2)^{3/4}}e^{-\frac{r^2}{2\sigma^2}} \,.

    :param width: Width of the atom-centered gaussian used to create the atomic density
    """

    def __init__(
        self,
        *,
        width: float,
        center_atom_weight: float = 1.0,
        scaling: Optional[RadialScaling] = None,
    ):
        super().__init__(center_atom_weight=center_atom_weight, scaling=scaling)
        self.width = float(width)

    def get_hypers(self):
        return {"type": "Gaussian", "width": self.width}

    def compute(self, positions: np.ndarray, *, derivative: bool) -> np.ndarray:
        width_sq = self.width**2
        x = positions**2 / (2 * width_sq)

        density = np.exp(-x) / (np.pi * width_sq) ** (3 / 4)

        if derivative:
            density *= -positions / width_sq

        return density


class SmearedPowerLaw(AtomicDensity):
    r"""Smeared power law density, as used in LODE.

    This is a smooth, differentiable density that behaves like :math:`1 / r^p` as
    :math:`r` goes to infinity.

    It is defined as

    .. math::

        g(r) = \frac{1}{\Gamma\left(\frac{p}{2}\right)}
               \frac{\gamma\left( \frac{p}{2}, \frac{r^2}{2\sigma^2} \right)}
                    {r^p},

    where :math:`p` is the potential exponent, :math:`\Gamma(z)` is the Gamma function
    and :math:`\gamma(a, x)` is the incomplete lower Gamma function.

    For more information about the derivation of this density, see
    https://doi.org/10.1021/acs.jpclett.3c02375 and section D of the supplementary
    information.

    :param smearing: Smearing used to remove the singularity at 0 (:math:`\sigma` above)
    :param exponent: Potential exponent of the decorated atom density (:math:`p` above)
    """

    def __init__(
        self,
        *,
        smearing: float,
        exponent: int,
        center_atom_weight: float = 1.0,
        scaling: Optional[RadialScaling] = None,
    ):
        super().__init__(center_atom_weight=center_atom_weight, scaling=scaling)
        self.smearing = float(smearing)
        self.exponent = int(exponent)

    def get_hypers(self):
        return {
            "type": "SmearedPowerLaw",
            "smearing": self.smearing,
            "exponent": self.exponent,
        }

    def compute(self, positions: np.ndarray, *, derivative: bool) -> np.ndarray:
        if not HAS_SCIPY:
            raise ValueError("SmearedPowerLaw requires scipy to be installed")

        if self.exponent == 0:
            proxy = Gaussian(width=self.smearing)
            return proxy.compute(positions=positions, derivative=derivative)
        else:
            smearing_sq = self.smearing**2
            a = self.exponent / 2
            x = positions**2 / (2 * smearing_sq)

            # Evaluating the formula above at :math:`r=0` is problematic because
            # :math:`g(r)` is of the form :math:`0/0`. For practical implementations, it
            # is thus more convenient to rewrite the density as
            #
            # .. math::
            #
            #   g(r) = \frac{1}{\Gamma(a)}\frac{1}{\left(2 \sigma^2\right)^a}
            #     \begin{cases}
            #         \frac{1}{a} - \frac{x}{a+1} + \frac{x^2}{2(a+2)}+\mathcal{O}(x^3)
            #             & x < 10^{-5} \\
            #         \frac{\gamma(a,x)}{x^a}
            #             & x \geq 10^{-5}
            #     \end{cases}
            #
            # where :math:`a = p/2`. It is convenient to use the expression for
            # sufficiently small :math:`x` since the relative weight of the first
            # neglected term is on the order of :math:`1/6x^3`. Therefore, the threshold
            # :math:`x = 10^{-5}` leads to relative errors on the order of the machine
            # epsilon.
            with warnings.catch_warnings():
                # Even though we use `np.where` to apply the `_compute_close_zero`
                # method for small `x`, the `_compute_far_zero` method will also
                # evaluated for small `x` and sending RuntimeWarnings. We filter these
                # warnings to avoid that these are presented to the user.
                warnings.filterwarnings("ignore", category=RuntimeWarning)
                density = np.where(
                    x < 1e-5,
                    self._compute_close_zero(a, x, derivative=derivative),
                    self._compute_far_zero(a, x, derivative=derivative),
                )

            density *= 1 / scipy.special.gamma(a) / (2 * smearing_sq) ** a

            # add inner derivative: ∂x/∂r
            if derivative:
                density *= positions / smearing_sq

            return density

    def _compute_close_zero(self, a: float, x: np.ndarray, derivative: bool):
        if derivative:
            return -1 / (a + 1) + x / (a + 2)
        else:
            return 1 / a - x / (a + 1) + x**2 / (2 * (a + 2))

    def _compute_far_zero(self, a: float, x: np.ndarray, derivative: bool):
        if derivative:
            return (
                np.exp(-x)
                - a * scipy.special.gamma(a) * scipy.special.gammainc(a, x) / x**a
            ) / x
        else:
            return scipy.special.gamma(a) * scipy.special.gammainc(a, x) / x**a



from setuptools import build_meta


get_requires_for_build_sdist = build_meta.get_requires_for_build_sdist
prepare_metadata_for_build_wheel = build_meta.prepare_metadata_for_build_wheel
build_wheel = build_meta.build_wheel
build_sdist = build_meta.build_sdist


def get_requires_for_build_wheel(config_settings=None):
    defaults = build_meta.get_requires_for_build_wheel(config_settings)
    return defaults + [
        "cmake",
        "metatensor-core >=0.1.20,<0.2.0",
    ]



import json
import re
from typing import Any, Dict

from . import _rascaline


class BadHyperParameters(Exception):
    pass


def convert_hypers(origin, representation=None, hypers=None):
    """Convert hyper-parameters from other software into the format used by featomic.

    :param origin: which software do the hyper-parameters come from? Valid values are:

        - ``"rascaline"`` for old rascaline format;
    :param representation: which representation are these hyper for? The meaning depend
        on the ``origin``:

        - for ``origin="rascaline"``, this is the name of the calculator class;
    :param hypers: the hyper parameter to convert. The type depend on the ``origin``:

        - for ``origin="rascaline"``, this should be a dictionary;

    :return: A string containing the code corresponding to the requested representation
        and hypers
    """
    if origin == "rascaline":
        if representation in [
            "SphericalExpansion",
            "SphericalExpansionByPair",
            "SoapPowerSpectrum",
        ]:
            hypers = _rascaline.convert_soap(hypers)
        elif representation == "SoapRadialSpectrum":
            hypers = _rascaline.convert_radial_spectrum(hypers)
        elif representation == "LodeSphericalExpansion":
            hypers = _rascaline.convert_lode(hypers)
        else:
            raise ValueError(
                "no hyper conversion exists for rascaline representation "
                f"'{representation}'"
            )

        hypers_dict = json.dumps(hypers, indent=4)
        hypers_dict = re.sub(r"\bnull\b", "None", hypers_dict)
        return f"{representation}(**{hypers_dict})"
    else:
        raise ValueError(f"no hyper conversion exists for {origin} software")


def hypers_to_json(hypers_dict: Dict[str, Any]):
    """
    Convert from class version of featomic hyper-parameters to the JSON version.

    The class version would contain something like ``{"cutoff": Cutoff(radius=3.4)}``,
    which this function transforms into ``{"cutoff": {"radius": 3.4", "smoothing":
    {"type": "Step"}}}``.
    """
    json = {}
    for key, value in hypers_dict.items():
        if hasattr(value, "_featomic_hypers"):
            value = value._featomic_hypers()

        if isinstance(value, dict):
            value = hypers_to_json(value)
        json[key] = value
    return json



def convert_soap(hypers):
    """convert from old style rascaline hypers for SOAP-related representations"""
    cleaned = {
        "cutoff": _process_cutoff(hypers),
        "density": _process_density(hypers),
    }

    max_angular = _get_or_error(hypers, "max_angular", "<root>")
    radial, spline_accuracy = _process_radial_basis(hypers)
    cleaned["basis"] = {
        "type": "TensorProduct",
        "max_angular": max_angular,
        "radial": radial,
    }
    if spline_accuracy is not None:
        if isinstance(spline_accuracy, float):
            cleaned["basis"]["spline_accuracy"] = spline_accuracy
        else:
            cleaned["basis"]["spline_accuracy"] = None

    return cleaned


def convert_radial_spectrum(hypers):
    """convert from old style rascaline hypers for SOAP radial spectrum"""
    cleaned = {
        "cutoff": _process_cutoff(hypers),
        "density": _process_density(hypers),
    }

    radial, spline_accuracy = _process_radial_basis(hypers)
    cleaned["basis"] = {"radial": radial}
    if spline_accuracy is not None:
        if isinstance(spline_accuracy, float):
            cleaned["basis"]["spline_accuracy"] = spline_accuracy
        else:
            cleaned["basis"]["spline_accuracy"] = None

    return cleaned


def convert_lode(hypers):
    """convert from old style rascaline hypers for LODE spherical expansion"""

    cleaned = {
        "density": _process_density(hypers),
    }

    max_angular = _get_or_error(hypers, "max_angular", "<root>")
    radial, spline_accuracy = _process_radial_basis(hypers, lode=True)
    cleaned["basis"] = {
        "type": "TensorProduct",
        "max_angular": max_angular,
        "radial": radial,
    }
    if spline_accuracy is not None:
        if isinstance(spline_accuracy, float):
            cleaned["basis"]["spline_accuracy"] = spline_accuracy
        else:
            cleaned["basis"]["spline_accuracy"] = None

    k_cutoff = hypers.get("k_cutoff")
    if k_cutoff is not None:
        cleaned["k_cutoff"] = k_cutoff

    return cleaned


def _process_cutoff(hypers):
    cutoff = {
        "radius": _get_or_error(hypers, "cutoff", "<root>"),
    }

    cutoff_fn = _get_or_error(hypers, "cutoff_function", "<root>")
    if "Step" in cutoff_fn:
        cutoff["smoothing"] = {"type": "Step"}
    if "ShiftedCosine" in cutoff_fn:
        width = _get_or_error(
            cutoff_fn["ShiftedCosine"], "width", "cutoff_function.ShiftedCosine"
        )
        cutoff["smoothing"] = {"type": "ShiftedCosine", "width": width}

    return cutoff


def _process_density(hypers):
    gaussian_width = _get_or_error(hypers, "atomic_gaussian_width", "<root>")
    center_weight = _get_or_error(hypers, "center_atom_weight", "<root>")
    exponent = hypers.get("potential_exponent")

    if exponent is None:
        density = {
            "type": "Gaussian",
            "width": gaussian_width,
        }
    else:
        density = {
            "type": "SmearedPowerLaw",
            "smearing": gaussian_width,
            "exponent": exponent,
        }

    if center_weight != 1.0:
        density["center_atom_weight"] = center_weight

    if "radial_scaling" in hypers:
        radial_scaling = hypers["radial_scaling"]
        if radial_scaling is None:
            pass

        if "None" in radial_scaling:
            pass

        if "Willatt2018" in radial_scaling:
            exponent = _get_or_error(
                radial_scaling["Willatt2018"], "exponent", "radial_scaling.Willatt2018"
            )
            rate = _get_or_error(
                radial_scaling["Willatt2018"], "rate", "radial_scaling.Willatt2018"
            )
            scale = _get_or_error(
                radial_scaling["Willatt2018"], "scale", "radial_scaling.Willatt2018"
            )

            density["scaling"] = {
                "type": "Willatt2018",
                "exponent": exponent,
                "rate": rate,
                "scale": scale,
            }

    return density


def _process_radial_basis(hypers, lode=False):
    spline_accuracy = None
    max_radial = _get_or_error(hypers, "max_radial", "<root>") - 1
    radial_basis = _get_or_error(hypers, "radial_basis", "<root>")

    if "Gto" in radial_basis:
        radial = {"type": "Gto", "max_radial": max_radial}

        if lode:
            cutoff = _get_or_error(hypers, "cutoff", "<root>") - 1
            radial["radius"] = cutoff

        gto_basis = radial_basis["Gto"]
        do_splines = gto_basis.get("splined_radial_integral", True)
        if do_splines:
            spline_accuracy = gto_basis.get("spline_accuracy")
        else:
            spline_accuracy = False

    elif "TabulatedRadialIntegral" in radial_basis:
        raise NotImplementedError("TabulatedRadialIntegral radial basis")

    return radial, spline_accuracy


def _get_or_error(hypers, name, path):
    from . import BadHyperParameters

    if name not in hypers:
        raise BadHyperParameters(f"missing {name} at {path} in hypers")

    return hypers.pop(name)



import ctypes
from typing import List, Optional, Union

from metatensor import Labels, TensorMap
from metatensor._c_api import c_uintptr_t, mts_tensormap_t

from ._c_api import (
    FEATOMIC_BUFFER_SIZE_ERROR,
    featomic_calculation_options_t,
    featomic_system_t,
)
from ._c_lib import _get_library
from .status import FeatomicError, _check_featomic_pointer
from .systems import IntoSystem, wrap_system


def _call_with_growing_buffer(callback, initial=1024):
    bufflen = initial

    while True:
        buffer = ctypes.create_string_buffer(bufflen)
        try:
            callback(buffer, bufflen)
            break
        except FeatomicError as e:
            if e.status == FEATOMIC_BUFFER_SIZE_ERROR:
                # grow the buffer and retry
                bufflen *= 2
            else:
                raise
    return buffer.value.decode("utf8")


def _convert_systems(systems):
    try:
        return (featomic_system_t * 1)(wrap_system(systems)._as_featomic_system_t())
    except TypeError:
        # try iterating over the systems
        return (featomic_system_t * len(systems))(
            *list(wrap_system(s)._as_featomic_system_t() for s in systems)
        )


def _options_to_c(
    gradients,
    use_native_system,
    selected_samples,
    selected_properties,
    selected_keys,
):
    if gradients is None:
        gradients = []

    if not isinstance(gradients, list):
        raise ValueError(
            f"`gradients` parameter must be a list of str, got a {type(gradients)}"
        )
    for parameter in gradients:
        if not isinstance(parameter, str):
            raise ValueError(
                "`gradients` parameter must be a list of str, got a "
                f"{type(parameter)} in the list"
            )

    c_gradients = ctypes.ARRAY(ctypes.c_char_p, len(gradients))()
    for i, parameter in enumerate(gradients):
        c_gradients[i] = parameter.encode("utf8")

    c_options = featomic_calculation_options_t()
    c_options.gradients = c_gradients
    c_options.gradients_count = c_gradients._length_
    c_options.use_native_system = bool(use_native_system)

    # store data to keep alive here
    c_options.__keepalive = {}

    if selected_samples is None:
        # nothing to do, all pointers are already NULL
        pass
    elif isinstance(selected_samples, Labels):
        selected_samples = selected_samples._as_mts_labels_t()
        c_options.selected_samples.subset = ctypes.pointer(selected_samples)
        c_options.__keepalive["selected_samples"] = selected_samples
    elif isinstance(selected_samples, TensorMap):
        c_options.selected_samples.predefined = selected_samples._ptr
    else:
        raise ValueError(
            "expected selected samples to be either an `metatensor.Labels` "
            "instance, or an got `metatensor.TensorMap` instance, got "
            f"{type(selected_samples)} instead"
        )

    if selected_properties is None:
        # nothing to do, all pointers are already NULL
        pass
    elif isinstance(selected_properties, Labels):
        selected_properties = selected_properties._as_mts_labels_t()
        c_options.selected_properties.subset = ctypes.pointer(selected_properties)
        c_options.__keepalive["selected_properties"] = selected_properties
    elif isinstance(selected_properties, TensorMap):
        c_options.selected_properties.predefined = selected_properties._ptr
    else:
        raise ValueError(
            "expected selected properties to be either an `metatensor.Labels` "
            "instance, or an got `metatensor.TensorMap` instance, got "
            f"{type(selected_properties)} instead"
        )

    if selected_keys is None:
        # nothing to do, all pointers are already NULL
        pass
    elif isinstance(selected_keys, Labels):
        selected_keys = selected_keys._as_mts_labels_t()
        c_options.selected_keys = ctypes.pointer(selected_keys)
        c_options.__keepalive["selected_keys"] = selected_keys
    return c_options


class CalculatorBase:
    """
    This is the base class for all calculators in featomic, providing the
    :py:meth:`CalculatorBase.compute` function.

    One can initialize a ``Calculator`` in two ways: either directly with the registered
    name and JSON parameter string (which are documented in the
    :ref:`userdoc-calculators`); or through one of the child class documented below.

    :param name: name used to register this calculator
    :param parameters: JSON parameter string for the calculator
    """

    def __init__(self, name: str, parameters: str):
        self._c_name = name
        self._lib = _get_library()
        self._as_parameter_ = self._lib.featomic_calculator(
            name.encode("utf8"), parameters.encode("utf8")
        )
        _check_featomic_pointer(self._as_parameter_)

        self._selected_samples = None
        self._selected_features = None

    def __del__(self):
        if hasattr(self, "_lib"):
            # if we failed to load the lib, don't double error by trying to call
            # ``self._lib.featomic_calculator_free``
            self._lib.featomic_calculator_free(self)
        self._as_parameter_ = 0

    @property
    def name(self) -> str:
        """name of this calculator"""
        return _call_with_growing_buffer(
            lambda buffer, bufflen: self._lib.featomic_calculator_name(
                self, buffer, bufflen
            )
        )

    @property
    def c_name(self) -> str:
        """name used to register & create this calculator"""
        return self._c_name

    @property
    def parameters(self):
        """parameters (formatted as JSON) used to create this calculator"""
        return _call_with_growing_buffer(
            lambda buffer, bufflen: self._lib.featomic_calculator_parameters(
                self, buffer, bufflen
            )
        )

    @property
    def cutoffs(self) -> List[float]:
        """all the radial cutoffs used by this calculator's neighbors lists"""

        cutoffs = ctypes.POINTER(ctypes.c_double)()
        cutoffs_count = c_uintptr_t()
        self._lib.featomic_calculator_cutoffs(self, cutoffs, cutoffs_count)

        result = []
        for i in range(cutoffs_count.value):
            result.append(cutoffs[i])

        return result

    def compute(
        self,
        systems: Union[IntoSystem, List[IntoSystem]],
        *,
        gradients: Optional[List[str]] = None,
        use_native_system: bool = True,
        selected_samples: Optional[Union[Labels, TensorMap]] = None,
        selected_properties: Optional[Union[Labels, TensorMap]] = None,
        selected_keys: Optional[Labels] = None,
    ) -> TensorMap:
        r"""Runs a calculation with this calculator on the given ``systems``.

        :param systems: single system or list of systems on which to run the
            calculation. The systems will automatically be wrapped into compatible
            classes (using :py:func:`featomic.systems.wrap_system`). Multiple types of
            systems are supported, see the documentation of
            :py:class:`featomic.IntoSystem` to get the full list.

        :param use_native_system: If ``True`` (this is the default), copy data from the
            ``systems`` into Rust ``SimpleSystem``. This can be a lot faster than having
            to cross the FFI boundary often when accessing the neighbor list. Otherwise
            the Python neighbor list is used.

        :param gradients: List of gradients to compute. If this is ``None`` or an empty
            list ``[]``, no gradients are computed. Gradients are stored inside the
            different blocks, and can be accessed with
            ``descriptor.block(...).gradient(<parameter>)``, where ``<parameter>`` is a
            string describing the gradients. The following gradients are available:

            - ``"positions"``, for gradients of the representation with respect to
              atomic positions, with fixed cell matrix parameters. Positions gradients
              are computed as

              .. math::
                  \frac{\partial \langle q \vert A_i \rangle}
                       {\partial \mathbf{r_j}}

              where :math:`\langle q \vert A_i \rangle` is the representation around
              atom :math:`i` and :math:`\mathbf{r_j}` is the position vector of the atom
              :math:`j`.

              **Note**: Position gradients of an atom are computed with respect to all
              other atoms within the representation. To recover the force one has to
              accumulate all pairs associated with atom :math:`i`.

            - ``"strain"``, for gradients of the representation with respect to strain.
              These gradients are typically used to compute the virial, and from there
              the pressure acting on a system. To compute them, we pretend that all the
              positions :math:`\mathbf r` and unit cell :math:`\mathbf H` have been
              scaled by a strain matrix :math:`\epsilon`:

              .. math::
                 \mathbf r &\rightarrow \mathbf r \left(\mathbb{1} + \epsilon \right)\\
                 \mathbf H &\rightarrow \mathbf H \left(\mathbb{1} + \epsilon \right)

              and then take the gradients of the representation with respect to this
              matrix:

              .. math::
                  \frac{\partial \langle q \vert A_i \rangle}
                       {\partial \mathbf{\epsilon}}

            - ``"cell"``, for gradients of the representation with respect to the
              system's cell parameters. These gradients are computed at fixed positions,
              and often not what you want when computing gradients explicitly (they are
              mainly used in ``featomic.torch`` to integrate with backward
              propagation). If you are trying to compute the virial or the stress, you
              should use ``"strain"`` gradients instead.

              .. math::
                  \left. \frac{\partial \langle q \vert A_i \rangle}
                       {\partial \mathbf{H}} \right |_\mathbf{r}

        :param selected_samples: Set of samples on which to run the calculation. Use
            ``None`` to run the calculation on all samples in the ``systems`` (this is
            the default).

            If ``selected_samples`` is an :py:class:`metatensor.TensorMap`, then the
            samples for each key will be used as-is when computing the representation.

            If ``selected_samples`` is a set of :py:class:`metatensor.Labels` containing
            the same variables as the default set of samples for this calculator, then
            only entries from the full set that also appear in ``selected_samples`` will
            be used.

            If ``selected_samples`` is a set of :py:class:`metatensor.Labels` containing
            a subset of the variables of the default set of samples, then only samples
            from the default set with the same values for these variables as one of the
            entries in ``selected_samples`` will be used.

        :param selected_properties: Set of properties to compute. Use ``None`` to run
            the calculation on all properties (this is the default).

            If ``selected_properties`` is an :py:class:`metatensor.TensorMap`, then the
            properties for each key will be used as-is when computing the
            representation.

            If ``selected_properties`` is a set of :py:class:`metatensor.Labels`
            containing the same variables as the default set of properties for this
            calculator, then only entries from the full set that also appear in
            ``selected_properties`` will be used.

            If ``selected_properties`` is a set of :py:class:`metatensor.Labels`
            containing a subset of the variables of the default set of properties, then
            only properties from the default set with the same values for these
            variables as one of the entries in ``selected_properties`` will be used.

        :param selected_keys: Selection for the keys to include in the output. If this
            is ``None``, the default set of keys (as determined by the calculator) will
            be used. Note that this default set of keys can depend on which systems we
            are running the calculation on.
        """

        c_systems = _convert_systems(systems)
        tensor_map_ptr = ctypes.POINTER(mts_tensormap_t)()

        c_options = _options_to_c(
            gradients=gradients,
            use_native_system=use_native_system,
            selected_samples=selected_samples,
            selected_properties=selected_properties,
            selected_keys=selected_keys,
        )
        self._lib.featomic_calculator_compute(
            self, tensor_map_ptr, c_systems, c_systems._length_, c_options
        )

        return TensorMap._from_ptr(tensor_map_ptr)



import os


_HERE = os.path.dirname(__file__)

cmake_prefix_path = os.path.realpath(os.path.join(_HERE, "lib", "cmake"))
"""
Path containing the CMake configuration files for the underlying C library
"""



"""
Module that stores the ClebschGordanReal class for computing and caching Clebsch
Gordan coefficients for use in CG combinations.
"""

import math
from typing import Dict, List

import numpy as np
import wigners

from . import _dispatch
from ._backend import (
    BACKEND_IS_METATENSOR_TORCH,
    Array,
    Device,
    DType,
    Labels,
    TensorBlock,
    TensorMap,
)


try:
    import torch
except ImportError:
    pass


UNKNOWN_ARRAY_TYPE = (
    "unknown array type, only numpy arrays and torch tensors are supported"
)


def calculate_cg_coefficients(
    lambda_max: int,
    cg_backend: str,
    arrays_backend: str,
    dtype: DType,
    device: Device,
) -> TensorMap:
    """
    Calculates the Clebsch-Gordan coefficients for all possible combination of angular
    momenta up to ``lambda_max``.

    The structure of the returned :py:class:`TensorMap` depends on whether the backend
    used to perform CG tensor products uses sparse or dense operations.

    Dense:
        - samples: ``_``, i.e. a dummy sample.
        - components: ``[m1, m2, mu]`` on separate components axes, where ``m1`` and
          ``m2`` are the m component values for the two arrays being combined and ``mu``
          is the m component value for the resulting array.
        - properties: ``cg_coefficient = [0]``

    Sparse:
        - samples: ``(m1, m2, mu)``, where ``m1`` and ``m2`` are the m component values
          for the two arrays being combined and ``mu`` is the m component value for the
          resulting array.
        - components: ``[]``, i.e. no components axis.
        - properties: ``cg_coefficient = [0]``

    :param lambda_max: maximum angular momentum value to compute CG coefficients for.
    :param cg_backend: whether to use ``"python-spare"`` or ``"python-dense"`` format
        for storing the CG coefficients.
    :param arrays_backend: whether to use ``"numpy"`` or ``"torch"`` arrays to store the
        coefficients.
    :param dtype: the scalar type to use to store coefficients
    :param device: the computational device to use for calculations. This must be
        ``"cpu"`` if ``array_backend="numpy"``.
    :returns: :py:class:`TensorMap` of the Clebsch-Gordan coefficients.
    """
    # Build some 'like' arrays for dispatch
    if arrays_backend == "torch":
        if dtype == torch.float32:
            complex_dtype = torch.complex64
        elif dtype == torch.float64:
            complex_dtype = torch.complex128
        else:
            raise ValueError(
                f"invalid dtype ({dtype}), only torch.float32 and torch.float64 are "
                "supported"
            )

        complex_like = torch.empty(0, dtype=complex_dtype, device=device)
        real_like = torch.empty(0, dtype=dtype, device=device)
        if BACKEND_IS_METATENSOR_TORCH:
            labels_values_like = torch.empty(0, dtype=torch.int32, device=device)
        else:
            # we are using metatensor-core with torch arrays
            labels_values_like = np.empty(0, dtype=np.int32)
    else:
        assert arrays_backend == "numpy"
        if dtype == np.float32:
            complex_dtype = np.complex64
        elif dtype == np.float64:
            complex_dtype = np.complex128
        else:
            raise ValueError(
                f"invalid dtype ({dtype}), only np.float32 and np.float64 are supported"
            )

        complex_like = np.empty(0, dtype=complex_dtype)
        real_like = np.empty(0, dtype=dtype)
        labels_values_like = np.empty(0, dtype=np.int32)

    # Calculate the CG coefficients, stored as a dict of dense arrays. This is the
    # starting point for the conversion to a TensorMap of different formats depending on
    # the backend options.
    cg_coeff_dict = _build_dense_cg_coeff_dict(
        lambda_max,
        complex_like,
        labels_values_like,
        arrays_backend=arrays_backend,
        dtype=dtype,
        device=device,
    )

    # Build the CG cache depending on whether the CG backend is sparse or dense. The
    # dispatching of the arrays backends are accounted for by `real_like` and
    # `labels_values_like`.
    if cg_backend == "python-sparse":
        return _cg_coeff_dict_to_tensormap_sparse(
            cg_coeff_dict, real_like, labels_values_like
        )
    elif cg_backend == "python-dense":
        return _cg_coeff_dict_to_tensormap_dense(
            cg_coeff_dict, real_like, labels_values_like
        )
    else:
        raise ValueError(
            f"invalid `cg_backend`, got '{cg_backend}', "
            "only 'python-dense', or 'python-sparse' are supported"
        )


def _build_dense_cg_coeff_dict(
    lambda_max: int,
    complex_like: Array,
    labels_values_like: Array,
    arrays_backend: str,
    dtype: DType,
    device: Device,
) -> Dict[int, Array]:
    """
    Calculates the CG coefficients and stores them as dense arrays in a dictionary.

    Each dictionary entry is a dense array with shape (2 * l1 + 1, 2 * l2 + 1, 2 *
    lambda + 1).

    {
        (l1, l2, lambda):
            array(
                cg_{m1, m2, mu}^{l1, l2, lambda} ... for m1 in range(-l1, l1 + 1), for
                m2 in range(-l2, l2 + 1), for mu in range(-lambda, lambda + 1),

                shape=(2 * l1 + 1, 2 * l2 + 1, 2 * lambda + 1),
            )
        ... for l1 in range(0, l1_list) for l2 in range(0, l2_list) for lambda in
        range(0, range(|l1 - l2|, ..., |l1 + l2|))
    }

    where `cg_{m1, m2, mu}^{l1, l2, lambda}` is the Clebsch-Gordan coefficient that
    describes the combination of the `m1` irreducible component of the `l1` angular
    channel and the `m2` irreducible component of the `l2` angular channel into the
    irreducible tensor of order `lambda`. In all cases, these correspond to the non-zero
    CG coefficients, i.e. those in the range |-l, ..., +l| for each angular order l in
    {l1, l2, lambda}.

    :param lambda_max: maximum angular momentum  value to compute CG coefficients for.
    :param complex_like: an empty array of dtype complex, used for dispatching
        operations
    :param real_like: an empty array of floating points data, used for dispatching
        operations
    :param labels_values_like: an empty array of dtype int32, used for dispatching
        operations

    :returns: dictionary of dense CG coefficients.
    """
    # real-to-complex and complex-to-real transformations as matrices
    r2c: Dict[int, Array] = {}
    c2r: Dict[int, Array] = {}
    coeff_dict = {}

    for o3_lambda in range(0, lambda_max + 1):
        c2r[o3_lambda] = _complex2real(o3_lambda, like=complex_like)
        r2c[o3_lambda] = _real2complex(o3_lambda, like=complex_like)

    for l1 in range(lambda_max + 1):
        for l2 in range(lambda_max + 1):
            for o3_lambda in range(
                max(l1, l2) - min(l1, l2), min(lambda_max, (l1 + l2)) + 1
            ):
                complex_cg = _dispatch.to(
                    wigners.clebsch_gordan_array(l1, l2, o3_lambda),
                    backend=arrays_backend,
                    device=device,
                    dtype=complex_like.dtype,
                )
                # We want to create a CG coefficient for a real representation, using
                # the expression:
                #
                #  C_{l1m1l2m2}^{lm}
                #       = \sum_{m'm1'm2'} ...
                #             ... U_{mm'}^{l} C_{l1m1'm2m2'}^{lm'} ...
                #             ... U^{\dagger}_{m1'm1}^{l1} U^{\dagger}_{m2'm2}^{l2}
                #
                # where:
                #       U is the "c2r" transformation matrix below
                #       U^{\dagger} is the "r2c" transformation matrix below
                #
                # 0) We have a CG coefficient of shape:
                #       complex_cg.shape = (2*l1+1, 2*l2+1, 2*L+1)
                #    and transormation matrices of shape:
                #       c2r[L].shape = (2*L+1, 2*L+1)
                #       r2c[L].shape = (2*L+1, 2*L+1)
                #
                # 1) take the matrix product:
                #       \sum_{m1'} C_{l1m1'm2m2'}^{lm'} U^{\dagger}_{m1'm1}^{l1}
                #    this requires some permuting of axes of the objects involved to
                #    ensure proper matching for matmul. i.e. permute axes of the complex
                #    CG coefficient from: (m1, m2, m) -> (m, m2, m1)
                first_step = _dispatch.permute(complex_cg, [2, 1, 0]) @ r2c[l1]
                # The result `first_step` has shape (2*L+1, 2*l2+1, 2*l1+1)
                #
                # 2) take the matrix product:
                #       \sum_{m2'} ...
                #          ... (first_step)_{m',m1,m2'}^{l,l1} U^{\dagger}_{m2'm2}^{l2}
                first_step_swap = first_step.swapaxes(1, 2)
                # The result `first_step_swap` has shape (2*L+1, 2*l1+1, 2*l2+1)
                second_step = first_step_swap @ r2c[l2]
                # The result `second_step` has shape (2*L+1, 2*l1+1, 2*l2+1)
                #
                # 3) take the matrix product:
                #       \sum_{m'} U_{mm'}^{l} (second_step)_{m',m1,m2}^{l,l1,l2}
                second_step_swap = _dispatch.permute(second_step, [1, 0, 2])
                # The result `second_step_swap` has shape (2*l1+1, 2*L+1, 2*l2+1)
                third_step = c2r[o3_lambda] @ second_step_swap
                # The result `third_step` has shape (2*l1+1, 2*L+1, 2*l2+1)
                third_step_swap = _dispatch.permute(third_step, [0, 2, 1])
                # The result `third_step_swap` has shape (2*l1+1, 2*l2+1, 2*L+1)

                if (l1 + l2 + o3_lambda) % 2 == 0:
                    cg_l1l2lam_dense = _dispatch.real(third_step_swap)
                else:
                    cg_l1l2lam_dense = _dispatch.imag(third_step_swap)

                coeff_dict[(l1, l2, o3_lambda)] = _dispatch.to(
                    cg_l1l2lam_dense,
                    dtype=dtype,
                    device=device,
                )

    return coeff_dict


def _cg_coeff_dict_to_tensormap_dense(
    coeff_dict: Dict, real_like: Array, labels_values_like: Array
) -> TensorMap:
    """
    Converts the dictionary of dense CG coefficients coefficients to
    :py:class:`TensorMap` format, specifically for performing CG tensor products with
    the "python-dense" backend.
    """
    keys = Labels(
        ["l1", "l2", "lambda"],
        _dispatch.int_array_like(list(coeff_dict.keys()), labels_values_like),
    )
    blocks = []

    for l1l2lam_values in coeff_dict.values():
        # extending shape by samples and properties
        block_value_shape = (1,) + l1l2lam_values.shape + (1,)
        blocks.append(
            TensorBlock(
                values=_dispatch.contiguous(l1l2lam_values.reshape(block_value_shape)),
                samples=Labels(
                    ["_"],
                    _dispatch.int_array_like([0], labels_values_like).reshape(-1, 1),
                ),
                components=[
                    Labels(
                        ["m1"],
                        _dispatch.int_range_like(
                            0, l1l2lam_values.shape[0], labels_values_like
                        ).reshape(-1, 1),
                    ),
                    Labels(
                        ["m2"],
                        _dispatch.int_range_like(
                            0, l1l2lam_values.shape[1], labels_values_like
                        ).reshape(-1, 1),
                    ),
                    Labels(
                        ["mu"],
                        _dispatch.int_range_like(
                            0, l1l2lam_values.shape[2], labels_values_like
                        ).reshape(-1, 1),
                    ),
                ],
                properties=Labels(
                    ["cg_coefficient"],
                    _dispatch.int_array_like([0], labels_values_like).reshape(-1, 1),
                ),
            )
        )

    return TensorMap(keys, blocks)


def _cg_coeff_dict_to_tensormap_sparse(
    coeff_dict: Dict, real_like: Array, labels_values_like: Array
) -> TensorMap:
    """
    Converts the dictionary of dense CG coefficients coefficients to
    :py:class:`TensorMap` format, specifically for performing CG tensor products with
    the "python-sparse" backend.
    """
    dict_keys = list(coeff_dict.keys())
    keys = Labels(
        ["l1", "l2", "lambda"],
        _dispatch.int_array_like(dict_keys, labels_values_like),
    )
    blocks = []

    # For each (l1, l2, lambda) combination, build a TensorBlock of non-zero CG coeffs
    for l1, l2, o3_lambda in dict_keys:
        cg_l1l2lam_dense = coeff_dict[(l1, l2, o3_lambda)]

        # Find the dense indices of the non-zero CG coeffs
        nonzeros_cg_coeffs_idx = _dispatch.where(
            _dispatch.abs(cg_l1l2lam_dense) > 1e-15
        )

        # Create a sparse dictionary indexed by  of the non-zero CG coeffs
        cg_l1l2lam_sparse = {}
        for i in range(len(nonzeros_cg_coeffs_idx[0])):
            m1 = nonzeros_cg_coeffs_idx[0][i]
            m2 = nonzeros_cg_coeffs_idx[1][i]
            mu = nonzeros_cg_coeffs_idx[2][i]
            cg_l1l2lam_sparse[(m1, m2, mu)] = cg_l1l2lam_dense[m1, m2, mu]

        l1l2lam_sample_values = []
        for m1m2mu_key in cg_l1l2lam_sparse.keys():
            l1l2lam_sample_values.append(m1m2mu_key)
        # extending shape by samples and properties
        values = _dispatch.real_array_like(
            [*cg_l1l2lam_sparse.values()], real_like
        ).reshape(-1, 1)
        l1l2lam_sample_values = _dispatch.int_array_like(
            l1l2lam_sample_values, labels_values_like
        )
        # we have to move put the m1 m2 m3 inside a block so we can access it easier
        # inside cg combine function,
        blocks.append(
            TensorBlock(
                values=_dispatch.contiguous(values),
                samples=Labels(["m1", "m2", "mu"], l1l2lam_sample_values),
                components=[],
                properties=Labels(
                    ["cg_coefficient"],
                    _dispatch.int_array_like([0], labels_values_like).reshape(-1, 1),
                ),
            )
        )

    return TensorMap(keys, blocks)


# ============================
# ===== Helper functions
# ============================


def _real2complex(o3_lambda: int, like: Array) -> Array:
    """
    Computes a matrix that can be used to convert from real to complex-valued spherical
    harmonics(coefficients) of order ``o3_lambda``.

    This is meant to be applied to the left:
    ``real2complex @ [-o3_lambda, ..., +o3_lambda]``.

    See https://en.wikipedia.org/wiki/Spherical_harmonics#Real_form for details on the
    convention for how these transformations are defined.

    Operations are dispatched to the corresponding array type given by ``like``
    """
    result = _dispatch.zeros_like(like, shape=(2 * o3_lambda + 1, 2 * o3_lambda + 1))
    inv_sqrt_2 = 1.0 / math.sqrt(2.0)
    i_sqrt_2 = 1.0j / complex(math.sqrt(2.0))

    for m in range(-o3_lambda, o3_lambda + 1):
        if m < 0:
            # Positive part
            result[o3_lambda + m, o3_lambda + m] = -i_sqrt_2
            # Negative part
            result[o3_lambda + m, o3_lambda - m] = i_sqrt_2 * ((-1) ** m)

        if m == 0:
            result[o3_lambda, o3_lambda] = 1.0

        if m > 0:
            # Negative part
            result[o3_lambda + m, o3_lambda - m] = inv_sqrt_2
            # Positive part
            result[o3_lambda + m, o3_lambda + m] = inv_sqrt_2 * ((-1) ** m)

    return result.T


def _complex2real(o3_lambda: int, like) -> Array:
    """
    Converts from complex to real spherical harmonics. This is just given by the
    conjugate transpose of the real->complex transformation matrices.

    Operations are dispatched to the corresponding array type given by ``like``
    """
    return _dispatch.conjugate(_real2complex(o3_lambda, like)).T


# ================================================================ #
# =============== Functions for performing CG coupling =========== #
# ================================================================ #


def cg_couple(
    array: Array,
    o3_lambdas: List[int],
    cg_coefficients: TensorMap,
    cg_backend: str,
) -> List[Array]:
    """
    Go from an uncoupled product basis that behave like a product of spherical harmonics
    to a coupled basis that behaves like a single spherical harmonic.

    The ``array`` shape should be ``(n_s, 2 * l1 + 1, 2 * l2 + 1, n_q)``.
    ``n_s`` is the number of samples, and ``n_q`` is the number of properties.

    This function will output a list of arrays, whose shape will be ``[n_s, (2 *
    o3_lambda+1), n_q]``, with the requested ``o3_lambdas``.

    These arrays will contain the result of projecting from a product of spherical
    harmonic with degree ``l1`` and ``l2`` to a single spherical harmonic of degree
    ``o3_lambdas``; using Clebsch-Gordan coefficients. The values for ``l1`` and ``l2``
    are inferred from the size of the middle axis of the input array.

    The Clebsch-Gordan coefficients are cached in ``cg_coefficients``. These must be
    computed by :py:func:`calculate_cg_coefficients`.

    The operation is dispatched such that numpy arrays or torch tensors are
    automatically handled.

    :param array: input array with shape ``[n_s, 2 * l1 + 1, 2 * l2 + 1, n_q]``
    :param o3_lambdas: list of degrees of spherical harmonics to compute
    :param cg_coefficients: CG coefficients as returned by
        :py:func:`calculate_cg_coefficients` with the same ``cg_backed`` given to this
        function
    :param cg_backend: specifies the backend to use for the calculation. It can be
        ``"python-dense"``, or ``"python-sparse"``.

    :return: A list of array, one for each ``o3_lambda``
    """

    assert len(array.shape) == 4

    l1 = (array.shape[1] - 1) // 2
    l2 = (array.shape[2] - 1) // 2

    if cg_backend == "python-sparse":
        arrays = {}
        for m1 in range(2 * l1 + 1):
            for m2 in range(2 * l2 + 1):
                arrays[str((m1, m2))] = array[:, m1, m2, :]

        return [
            _cg_couple_sparse(arrays, l1, l2, o3_lambda, cg_coefficients)
            for o3_lambda in o3_lambdas
        ]
    elif cg_backend == "python-dense":
        n_samples = array.shape[0]
        n_properties = array.shape[3]

        # Move properties next to samples
        array = _dispatch.permute(array, [0, 3, 1, 2])

        array = array.reshape(n_samples * n_properties, 2 * l1 + 1, 2 * l2 + 1)

        results = []
        for o3_lambda in o3_lambdas:
            result = _cg_couple_dense(array, o3_lambda, cg_coefficients)

            # Separate samples and properties
            result = result.reshape(n_samples, n_properties, -1)

            # Back to TensorBlock-like axes
            result = result.swapaxes(1, 2)

            results.append(result)

        return results

    else:
        raise ValueError(
            f"invalid `cg_backend`, got '{cg_backend}', "
            "only 'python-dense', or 'python-sparse' are supported"
        )


def _cg_couple_sparse(
    arrays: Dict[str, Array],
    l1: int,
    l2: int,
    o3_lambda: int,
    cg_coefficients: TensorMap,
):
    """
    Couple two spherical harmonics (of degree ``l1`` and ``l2``) to a single one (of
    degree ``o3_lambda``) using CG coefficients. This is a "dense" implementation, using
    all CG coefficients at the same time.

    :param l1: degree of the first spherical harmonic
    :param l2: degree of the first spherical harmonic
    :param arrays: dictionary of arrays for the different ``(m1, m2)`` corresponding to
        ``l1`` and ``l2``. Each array in the dictionary is expected to have a shape
        ``[samples, properties]`` and correspond to a single ``(m1, m2)`` pair.
    :param o3_lambda: value of lambda for the output spherical harmonic
    :param cg_coefficients: CG coefficients as returned by
        :py:func:`calculate_cg_coefficients` with ``cg_backed="python-sparse"``
    """

    array = list(arrays.values())[0]
    assert len(array.shape) == 2

    output = _dispatch.zeros_like(
        array, (2 * o3_lambda + 1, array.shape[0], array.shape[1])
    )

    # Fill in each mu component of the output array in turn
    cg_l1l2lam = cg_coefficients.block({"l1": l1, "l2": l2, "lambda": o3_lambda})
    cg_l1l2lam_samples = cg_l1l2lam.samples
    for i in range(len(cg_l1l2lam_samples)):
        m1m2mu = cg_l1l2lam_samples.entry(i)
        m1 = int(m1m2mu[0])
        m2 = int(m1m2mu[1])
        mu = int(m1m2mu[2])
        # Broadcast arrays, multiply together and with CG coeff
        output[mu, :, :] += (
            arrays[str((m1, m2))]
            * (-1) ** (l1 + l2 + o3_lambda)
            * cg_l1l2lam.values[i, 0]
        )

    return output.swapaxes(0, 1)


def _cg_couple_dense(
    array: Array,
    o3_lambda: int,
    cg_coefficients: TensorMap,
) -> Array:
    """
    Couple two spherical harmonics (of degree ``l1`` and ``l2``) to a single one (of
    degree ``o3_lambda``) using CG coefficients. This is a "dense" implementation, using
    all CG coefficients at the same time.

    :param array: input array, we expect a shape of ``[N, 2*l1 + 1, 2*l2 + 1]``
    :param o3_lambda: value of lambda for the output spherical harmonic
    :param cg_coefficients: CG coefficients as returned by
        :py:func:`calculate_cg_coefficients` with ``cg_backed="python-dense"``

    :return: array of shape ``[N, 2 * o3_lambda + 1]``
    """
    assert len(array.shape) == 3

    l1 = (array.shape[1] - 1) // 2
    l2 = (array.shape[2] - 1) // 2

    cg_l1l2lam = (-1) ** (l1 + l2 + o3_lambda) * cg_coefficients.block(
        {"l1": l1, "l2": l2, "lambda": o3_lambda}
    ).values

    return _dispatch.tensordot(array, cg_l1l2lam[0, ..., 0], axes=([2, 1], [1, 0]))


# ======================================================================= #
# =============== Functions for performing CG tensor products =========== #
# ======================================================================= #


def cg_tensor_product(
    array_1: Array,
    array_2: Array,
    o3_lambdas: List[int],
    cg_coefficients: TensorMap,
    cg_backend: str,
) -> List[Array]:
    """
    Compute the Clebsch-Gordan tensor product of ``array_1`` and ``array_2``.

    ``array_1`` shape should be ``(n_samples, 2 * l1 + 1, n_q)`` and ``array_2`` shape
    should be ``(n_samples, 2 * l2 + 1, n_p)``. ``n_samples`` is the number of samples,
    ``n_q`` and ``n_p`` are the number of properties in each array. The number of
    samples in both array must be the same.

    This function will output a list of arrays, whose shape will be ``[n_samples, (2 *
    o3_lambda+1), n_q * n_p]``, with the requested ``o3_lambdas``. These arrays will
    contain the result of computing a tensor product of ``array_1`` and ``array_2``,
    followed by a projection from the product of spherical harmonic with degree ``l1``
    and ``l2`` to a single spherical harmonic of degree ``o3_lambdas``; using
    Clebsch-Gordan coefficients. The values for ``l1`` and ``l2`` are inferred from the
    size of the middle axis of the input arrays.

    The Clebsch-Gordan coefficients are cached in ``cg_coefficients``. These must be
    computed by :py:func:`calculate_cg_coefficients`.

    The operation is dispatched such that numpy arrays or torch tensors are
    automatically handled.

    ``cg_backend="metadata"`` can be used to return an empty array of the correct shape,
    without performing any calculation. This can be useful for probing the outputs of CG
    iterations in terms of metadata without the computational cost of performing the CG
    combinations — i.e. using :py:func:`DensityCorrelations.compute_metadata`.

    :param array_1: array with shape ``[n_samples, 2 * l1 + 1, n_q]``
    :param array_2: array with shape ``[n_samples, 2 * l2 + 1, n_p]``
    :param o3_lambdas: list of degrees of spherical harmonics to compute
    :param cg_coefficients: a :py:class:`TensorMap` containing CG coefficients in a
        format for either sparse or dense CG tensor products, as returned by
        :py:func:`calculate_cg_coefficients`. Only used if ``cg_backend`` is not
        ``"metadata"``.
    :param cg_backend: specifies the backend to use for the calculation. It can be
        ``"python-dense"``, ``"python-sparse"``, or ``"metadata"``. If
        ``"python-dense"`` or ``"python-sparse"`` is chosen, a dense or sparse
        combination (respectively) of the arrays is performed. If ``"metadata"`` is
        chosen, no combination is performed, and an empty array of the correct shape is
        returned instead.

    :returns: list of arrays of shape ``[n_samples, (2*o3_lambda+1), n_q * n_p]``
    """
    # If just precomputing metadata, return an empty array
    if cg_backend == "metadata":
        return [
            _cg_tensor_product_empty(array_1, array_2, o3_lambda)
            for o3_lambda in o3_lambdas
        ]

    elif cg_backend == "python-sparse":
        return _cg_tensor_product_sparse(array_1, array_2, o3_lambdas, cg_coefficients)
    elif cg_backend == "python-dense":
        return _cg_tensor_product_dense(array_1, array_2, o3_lambdas, cg_coefficients)
    else:
        raise ValueError(
            f"invalid `cg_backend`, got '{cg_backend}', "
            "only 'python-dense', 'python-sparse', or 'metadata' are supported"
        )


def _cg_tensor_product_empty(array_1: Array, array_2: Array, o3_lambda: int) -> Array:
    """
    Returns an empty array of the correct shape, imitating the output array shape
    produced by a CG combination of ``array_1`` and ``array_2``.
    """
    # Samples dimensions must be the same
    assert array_1.shape[0] == array_2.shape[0]

    # Define other useful dimensions
    n_s = array_1.shape[0]  # number of samples
    n_p = array_1.shape[2]  # number of properties in array_1
    n_q = array_2.shape[2]  # number of properties in array_2

    return _dispatch.empty_like(array_1, (n_s, 2 * o3_lambda + 1, n_p * n_q))


def _cg_tensor_product_sparse(
    array_1: Array,
    array_2: Array,
    o3_lambdas: List[int],
    cg_coefficients: TensorMap,
) -> List[Array]:
    """compute the Clebsch-Gordan tensor product of 2 arrays using sparse operations"""
    # Samples dimensions must be the same
    assert array_1.shape[0] == array_2.shape[0]

    # Infer l1 and l2 from the len of the length of axis 1 of each tensor
    l1 = (array_1.shape[1] - 1) // 2
    l2 = (array_2.shape[1] - 1) // 2

    # Define other useful dimensions
    n_s = array_1.shape[0]  # number of samples
    n_p = array_1.shape[2]  # number of properties in array_1
    n_q = array_2.shape[2]  # number of properties in array_2

    # Compute the partial tensor products
    partial_tensor_products = {}
    for o3_lambda in o3_lambdas:
        cg_l1l2lam = cg_coefficients.block({"l1": l1, "l2": l2, "lambda": o3_lambda})
        for i in range(len(cg_l1l2lam.samples)):
            m1m2mu = cg_l1l2lam.samples.entry(i)
            m1 = int(m1m2mu[0])
            m2 = int(m1m2mu[1])

            # We use a string as dict key since TorchScript does not support (int, int)
            dict_key = str((m1, m2))
            if dict_key not in partial_tensor_products:
                partial_tensor_products[dict_key] = (
                    array_1[:, m1, :, None] * array_2[:, m2, None, :]
                ).reshape(n_s, n_p * n_q)

    # Couple l1,l2 to the different lambda, re-using the partial tensor products
    result = []
    for o3_lambda in o3_lambdas:
        output = _cg_couple_sparse(
            partial_tensor_products, l1, l2, o3_lambda, cg_coefficients
        )
        result.append(output)

    return result


def _cg_tensor_product_dense(
    array_1: Array,
    array_2: Array,
    o3_lambdas: List[int],
    cg_coefficients: TensorMap,
) -> List[Array]:
    """compute the Clebsch-Gordan tensor product of 2 arrays using dense operations"""
    # Infer l1 and l2 from the len of the length of axis 1 of each tensor
    l1 = (array_1.shape[1] - 1) // 2
    l2 = (array_2.shape[1] - 1) // 2

    # Define other useful dimensions
    n_s = array_1.shape[0]  # number of samples
    n_p = array_1.shape[2]  # number of properties in array_1
    n_q = array_2.shape[2]  # number of properties in array_2

    # we broadcast it in this way so we only need to do one swapaxes in the next step
    # The resulting shape of tensor_product is [samples, l2, p, l1, q]
    tensor_product = array_1[:, None, None, :, :] * array_2[:, :, :, None, None]

    # [samples, l2, p, l1, q] -> [samples, q, p, l1, l2]
    tensor_product = _dispatch.swapaxes(tensor_product, 1, 4)
    # [samples, l2, p, l1, q] -> [(samples q p), l1, l2]
    tensor_product = tensor_product.reshape(-1, 2 * l1 + 1, 2 * l2 + 1)

    result = []
    for o3_lambda in o3_lambdas:
        # output shape is [(samples q p), lambda]
        output = _cg_couple_dense(tensor_product, o3_lambda, cg_coefficients)
        #  => [samples, (q p), lambda]
        output = output.reshape(n_s, (n_p * n_q), 2 * o3_lambda + 1)
        #  => [samples, lambda, (q p)]
        output = _dispatch.swapaxes(output, 1, 2)
        result.append(output)

    return result



import json

from featomic._hypers import BadHyperParameters, convert_hypers, hypers_to_json


try:
    # see featomic-torch/calculators.py for the explanation of what's going on here
    _ = CalculatorBase
except NameError:
    from .calculator_base import CalculatorBase


class AtomicComposition(CalculatorBase):
    """An atomic composition calculator for obtaining the stoichiometric information.

    For ``per_system=False`` calculator has one property ``count`` that is ``1`` for all
    atoms, and has a sample index that indicates the central atom type.

    For ``per_system=True`` a sum for each system is performed. The number of atoms per
    system is saved. The only sample left is named ``system``.
    """

    def __init__(self, *, per_system):
        parameters = hypers_to_json(
            {
                "per_system": per_system,
            }
        )
        super().__init__("atomic_composition", json.dumps(parameters))


class DummyCalculator(CalculatorBase):
    def __init__(self, *, cutoff, delta, name):
        parameters = hypers_to_json(
            {
                "cutoff": cutoff,
                "delta": delta,
                "name": name,
            }
        )
        super().__init__("dummy_calculator", json.dumps(parameters))


class NeighborList(CalculatorBase):
    """
    This calculator computes the neighbor list for a given spherical cutoff, and returns
    the list of distance vectors between all pairs of atoms strictly inside the cutoff.

    Users can request either a "full" neighbor list (including an entry for both ``i -
    j`` pairs and ``j - i`` pairs) or save memory/computational by only working with
    "half" neighbor list (only including one entry for each ``i/j`` pair)

    Pairs between an atom and it's own periodic copy can appear when the cutoff is
    larger than the cell under periodic boundary conditions. Self pairs with a distance
    of 0 (i.e. self pairs inside the original unit cell) are only included when using
    ``self_pairs=True``.

    This calculator produces a single property (``"distance"``) with three components
    (``"pair_xyz"``) containing the x, y, and z component of the distance vector of the
    pair.

    The samples contain the two atoms indexes, as well as the number of cell boundaries
    crossed to create this pair.
    """

    def __init__(self, *, cutoff, full_neighbor_list, self_pairs=False):
        parameters = hypers_to_json(
            {
                "cutoff": cutoff,
                "full_neighbor_list": full_neighbor_list,
                "self_pairs": self_pairs,
            }
        )
        super().__init__("neighbor_list", json.dumps(parameters))


class SortedDistances(CalculatorBase):
    """Sorted distances vector representation of an atomic environment.

    Each atomic center is represented by a vector of distance to its neighbors
    within the spherical ``cutoff``, sorted from smallest to largest. If there
    are less neighbors than ``max_neighbors``, the remaining entries are filled
    with ``cutoff`` instead.

    Separate atomic types for neighbors are represented separately, meaning that the
    ``max_neighbors`` parameter only apply to a single atomic type.

    For a full description of the hyper-parameters, see the corresponding
    :ref:`documentation <sorted-distances>`.
    """

    def __init__(self, *, cutoff, max_neighbors, separate_neighbor_types):
        parameters = hypers_to_json(
            {
                "cutoff": cutoff,
                "max_neighbors": max_neighbors,
                "separate_neighbor_types": separate_neighbor_types,
            }
        )
        super().__init__("sorted_distances", json.dumps(parameters))


def _check_for_old_hypers(calculator, hypers):
    try:
        new_hypers = convert_hypers(
            origin="rascaline",
            representation=calculator,
            hypers=hypers,
        )
    except BadHyperParameters as e:
        print(e)
        raise ValueError(
            f"invalid hyper parameters to {calculator}, "
            "expected `density` and `basis` to be present"
        )

    raise ValueError(
        f"{calculator} hyper parameter changed recently, "
        "please update your code. Here are the new equivalent parameters:\n"
        + new_hypers
    )


class SphericalExpansion(CalculatorBase):
    """Spherical expansion of Smooth Overlap of Atomic Positions (SOAP).

    The spherical expansion is at the core of representations in the SOAP family
    of descriptors. The spherical expansion represent atomic density as a
    collection of Gaussian functions centered on each atom, and then represent
    the local density around each atom (up to a cutoff) on a basis of radial
    functions and spherical harmonics. This representation is not rotationally
    invariant, for that you should use the :py:class:`SoapPowerSpectrum` class.

    See `this review article <https://doi.org/10.1063/1.5090481>`_ for more
    information on the SOAP representation, and `this paper
    <https://doi.org/10.1063/5.0044689>`_ for information on how it is
    implemented in featomic.

    For a full description of the hyper-parameters, see the corresponding
    :ref:`documentation <spherical-expansion>`.
    """

    def __init__(self, *, cutoff=None, density=None, basis=None, **kwargs):
        if len(kwargs) != 0 or density is None or basis is None:
            _check_for_old_hypers("SphericalExpansion", {"cutoff": cutoff, **kwargs})

        parameters = hypers_to_json(
            {
                "cutoff": cutoff,
                "density": density,
                "basis": basis,
            }
        )

        super().__init__("spherical_expansion", json.dumps(parameters))


class SphericalExpansionByPair(CalculatorBase):
    """
    Pair-by-pair version of the spherical expansion of Smooth Overlap of Atomic
    Positions (SOAP).

    This is usually an intermediary step when computing the full
    :py:class:`SphericalExpansion`, which most users do not need to care about.
    When computing SOAP, the density around the central atom :math:`i` is a sum
    of pair contributions:

    .. math::
        \\rho_i(\\mathbf{r}) = \\sum_j g(\\mathbf{r_{ji}} - \\mathbf{r}).

    The full spherical expansion is then computed as a sum over all pairs within
    the cutoff:

    .. math::
        \\int Y^l_m(\\mathbf{r})\\ R_n(\\mathbf{r}) \\, \\rho_i(\\mathbf{r})
            \\mathrm{d}\\mathbf{r} = \\sum_j C^{ij}_{nlm}

        C^{ij}_{nlm} = \\int Y^l_m(\\mathbf{r}) \\ R_n(\\mathbf{r}) \\,
            g(\\mathbf{r_{ji}} - \\mathbf{r}) \\, \\mathrm{d}\\mathbf{r}

    This calculators computes the individual :math:`C^{ij}_{nlm}` terms, which
    can then be used to build multi-center representations such as the ones
    discussed in `"Unified theory of atom-centered representations and
    message-passing machine-learning schemes"
    <https://doi.org/10.1063/5.0087042>`_.

    For a full description of the hyper-parameters, see the corresponding
    :ref:`documentation <spherical-expansion-by-pair>`.
    """

    def __init__(self, *, cutoff=None, density=None, basis=None, **kwargs):
        if len(kwargs) != 0 or density is None or basis is None:
            _check_for_old_hypers(
                "SphericalExpansionByPair", {"cutoff": cutoff, **kwargs}
            )

        parameters = hypers_to_json(
            {
                "cutoff": cutoff,
                "density": density,
                "basis": basis,
            }
        )

        super().__init__("spherical_expansion_by_pair", json.dumps(parameters))


class SoapRadialSpectrum(CalculatorBase):
    """Radial spectrum of Smooth Overlap of Atomic Positions (SOAP).

    The SOAP radial spectrum represent each atom by the radial average of the
    density of its neighbors. It is very similar to a radial distribution
    function `g(r)`. It is a 2-body representation, only containing information
    about the distances between atoms.

    See `this review article <https://doi.org/10.1063/1.5090481>`_ for more
    information on the SOAP representation, and `this paper
    <https://doi.org/10.1063/5.0044689>`_ for information on how it is
    implemented in featomic.

    For a full description of the hyper-parameters, see the corresponding
    :ref:`documentation <soap-radial-spectrum>`.
    """

    def __init__(self, *, cutoff=None, density=None, basis=None, **kwargs):
        if len(kwargs) != 0 or density is None or basis is None:
            _check_for_old_hypers("SoapRadialSpectrum", {"cutoff": cutoff, **kwargs})

        parameters = hypers_to_json(
            {
                "cutoff": cutoff,
                "density": density,
                "basis": basis,
            }
        )

        super().__init__("soap_radial_spectrum", json.dumps(parameters))


class SoapPowerSpectrum(CalculatorBase):
    """Power spectrum of Smooth Overlap of Atomic Positions (SOAP).

    The SOAP power spectrum is the main member of the SOAP
    family of descriptors. The power spectrum is based on the
    :py:class:`SphericalExpansion` coefficients, which are combined to create a
    rotationally invariant three-body descriptor.

    See `this review article <https://doi.org/10.1063/1.5090481>`_ for more
    information on the SOAP representation, and `this paper
    <https://doi.org/10.1063/5.0044689>`_ for information on how it is
    implemented in featomic.

    For a full description of the hyper-parameters, see the corresponding
    :ref:`documentation <soap-power-spectrum>`.

    .. seealso::
        :py:class:`featomic.utils.PowerSpectrum` is an implementation that
        allows to compute the power spectrum from different spherical expansions.
    """

    def __init__(self, *, cutoff=None, density=None, basis=None, **kwargs):
        if len(kwargs) != 0 or density is None or basis is None:
            _check_for_old_hypers("SoapPowerSpectrum", {"cutoff": cutoff, **kwargs})

        parameters = hypers_to_json(
            {
                "cutoff": cutoff,
                "density": density,
                "basis": basis,
            }
        )

        super().__init__("soap_power_spectrum", json.dumps(parameters))


class LodeSphericalExpansion(CalculatorBase):
    """Long-Distance Equivariant (LODE).

    The spherical expansion is at the core of representations in the LODE
    family. The LODE spherical
    expansion represent atomic density as a collection of 'decorated' gaussian
    functions centered on each atom, and then represent the local density around
    each atom on a basis of radial functions and spherical harmonics.
    This representation is not rotationally invariant.

    See `this article <https://aip.scitation.org/doi/10.1063/1.5128375>`_
    for more information on the LODE representation.

    For a full description of the hyper-parameters, see the corresponding
    :ref:`documentation <lode-spherical-expansion>`.
    """

    def __init__(self, *, density=None, basis=None, k_cutoff=None, **kwargs):
        if len(kwargs) != 0 or density is None or basis is None:
            _check_for_old_hypers(
                "LodeSphericalExpansion", {"k_cutoff": k_cutoff, **kwargs}
            )

        parameters = hypers_to_json(
            {
                "k_cutoff": k_cutoff,
                "density": density,
                "basis": basis,
            }
        )

        super().__init__("lode_spherical_expansion", json.dumps(parameters))



from ._c_lib import _get_library
from .calculator_base import _call_with_growing_buffer


class Profiler:
    """Profiler recording execution time of featomic functions.

    Featomic uses the `time_graph <https://docs.rs/time-graph/>`_ to collect
    timing information on the calculations. The ``Profiler`` class can be used
    as a context manager to access to this functionality.

    The profiling code collects the total time spent inside the most important
    functions, as well as the function call graph (which function called which
    other function).

    .. code-block:: python

        import featomic

        with featomic.Profiler() as profiler:
            # run some calculations
            ...

        print(profiler.as_short_table())
    """

    def __init__(self):
        self._lib = _get_library()

    def __enter__(self):
        self._lib.featomic_profiling_enable(True)
        self._lib.featomic_profiling_clear()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._lib.featomic_profiling_enable(False)

    def as_json(self):
        """Get current profiling data formatted as JSON."""
        return _call_with_growing_buffer(
            lambda b, s: self._lib.featomic_profiling_get("json".encode("utf8"), b, s)
        )

    def as_table(self):
        """Get current profiling data formatted as a table."""
        return _call_with_growing_buffer(
            lambda b, s: self._lib.featomic_profiling_get("table".encode("utf8"), b, s)
        )

    def as_short_table(self):
        """
        Get current profiling data formatted as a table, using short functions names.
        """
        return _call_with_growing_buffer(
            lambda b, s: self._lib.featomic_profiling_get(
                "short_table".encode("utf8"), b, s
            )
        )



import logging
import warnings

from ._c_api import (
    FEATOMIC_LOG_LEVEL_DEBUG,
    FEATOMIC_LOG_LEVEL_ERROR,
    FEATOMIC_LOG_LEVEL_INFO,
    FEATOMIC_LOG_LEVEL_TRACE,
    FEATOMIC_LOG_LEVEL_WARN,
    featomic_logging_callback_t,
)


_CURRENT_CALLBACK = None


def default_logging_callback(level, message):
    """Redirect message to the ``logging`` module."""
    if level == FEATOMIC_LOG_LEVEL_ERROR:
        logging.error(message)
    elif level == FEATOMIC_LOG_LEVEL_WARN:
        logging.warning(message)
    elif level == FEATOMIC_LOG_LEVEL_INFO:
        logging.info(message)
    elif level == FEATOMIC_LOG_LEVEL_DEBUG:
        logging.debug(message)
    elif level == FEATOMIC_LOG_LEVEL_TRACE:
        logging.debug(message)
    else:
        raise ValueError(f"Log level {level} is not supported.")


def set_logging_callback(function):
    """Call ``function`` on every log event.

    The callback functions should take two arguments: an integer value
    representing the log level and a string containing the log message. The
    function return value is ignored.
    """
    from ._c_lib import _get_library

    library = _get_library()
    _set_logging_callback_impl(library, function)


def _set_logging_callback_impl(library, function):
    """Implementation of :py:func:`set_logging_callback`

    This function gets the :py:class:`ctypes.CDLL` instance for ``libfeatomic``
    as a parameter.

    This is used to be able to setup the default logging callback when loading
    the library, without a recursive call to :py:func:`_get_library` in this
    function.
    """

    def wrapper(log_level, message):
        try:
            function(log_level, message.decode("utf8"))
        except Exception as e:
            warnings.warn(
                message=f"exception raised in logging callback: {e}",
                category=ResourceWarning,
                stacklevel=1,
            )

    # store the current callback in a global python variable to prevent it from
    # being garbage-collected.
    global _CURRENT_CALLBACK
    _CURRENT_CALLBACK = featomic_logging_callback_t(wrapper)

    library.featomic_set_logging_callback(_CURRENT_CALLBACK)



from ._cartesian_spherical import cartesian_to_spherical  # noqa: F401
from ._cg_product import ClebschGordanProduct  # noqa: F401
from ._coefficients import calculate_cg_coefficients  # noqa: F401
from ._density_correlations import DensityCorrelations  # noqa: F401
from ._equivariant_power_spectrum import EquivariantPowerSpectrum  # noqa: F401
from ._equivariant_power_spectrum_by_pair import (  # noqa: F401
    EquivariantPowerSpectrumByPair,  # noqa: F401
)  # noqa: F401
from ._power_spectrum import PowerSpectrum  # noqa: F401



"""Helper functions to dispatch methods between numpy and torch.

The functions are similar to those in metatensor-operations. Missing functions may
already exist there. Functions are ordered alphabetically.
"""

import itertools
from typing import List, Optional, Union

import numpy as np

from ._backend import Device, DType, torch_jit_is_scripting


try:
    import torch
    from torch import Tensor as TorchTensor
except ImportError:

    class TorchTensor:
        pass


UNKNOWN_ARRAY_TYPE = (
    "unknown array type, only numpy arrays and torch tensors are supported"
)


def _check_all_torch_tensor(arrays: List[TorchTensor]):
    for array in arrays:
        if not isinstance(array, TorchTensor):
            raise TypeError(
                f"expected argument to be a torch.Tensor, but got {type(array)}"
            )


def _check_all_np_ndarray(arrays):
    for array in arrays:
        if not isinstance(array, np.ndarray):
            raise TypeError(
                f"expected argument to be a np.ndarray, but got {type(array)}"
            )


def concatenate(arrays: List[TorchTensor], axis: int):
    """
    Concatenate a group of arrays along a given axis.

    This function has the same behavior as ``numpy.concatenate(arrays, axis)``
    and ``torch.concatenate(arrays, axis)``.

    Passing `axis` as ``0`` is equivalent to :py:func:`numpy.vstack`, ``1`` to
    :py:func:`numpy.hstack`, and ``2`` to :py:func:`numpy.dstack`, though any
    axis index > 0 is valid.
    """
    if isinstance(arrays[0], TorchTensor):
        _check_all_torch_tensor(arrays)
        return torch.concatenate(arrays, axis)
    elif isinstance(arrays[0], np.ndarray):
        _check_all_np_ndarray(arrays)
        return np.concatenate(arrays, axis)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def permute(array, axes: List[int]):
    """
    Permute the axes of the given ``array``. For numpy and torch, ``transpose`` and
    ``permute`` are equivalent operations.
    """
    if isinstance(array, TorchTensor):
        return torch.permute(array, axes)
    elif isinstance(array, np.ndarray):
        return np.transpose(array, axes)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def tensordot(array_1, array_2, axes: List[List[int]]):
    """
    Compute tensor dot product along specified axes for arrays.

    This function has the same behavior as ``np.tensordot(array_1, array_2, axes)`` and
    ``torch.tensordot(array_1, array_2, axes)``.
    """
    if isinstance(array_1, TorchTensor):
        return torch.tensordot(array_1, array_2, axes)
    elif isinstance(array_1, np.ndarray):
        return np.tensordot(array_1, array_2, axes)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def empty_like(array, shape: Optional[List[int]] = None, requires_grad: bool = False):
    """
    Create an uninitialized array, with the given ``shape``, and similar dtype, device
    and other options as ``array``.

    If ``shape`` is ``None``, the array shape is used instead. ``requires_grad`` is only
    used for torch tensors, and set the corresponding value on the returned array.

    This is the equivalent to ``np.empty_like(array, shape=shape)``.
    """
    if isinstance(array, TorchTensor):
        if shape is None:
            shape = array.size()
        return torch.empty(
            shape,
            dtype=array.dtype,
            layout=array.layout,
            device=array.device,
        ).requires_grad_(requires_grad)
    elif isinstance(array, np.ndarray):
        return np.empty_like(array, shape=shape, subok=False)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def list_to_array(array, data: List[List[int]]):
    """Create an object from data with the same type as ``array``."""
    if isinstance(array, TorchTensor):
        return torch.tensor(data)
    elif isinstance(array, np.ndarray):
        return np.array(data)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def matmul(a, b):
    """Matrix product of two arrays."""
    if isinstance(a, TorchTensor):
        _check_all_torch_tensor([b])
        return torch.matmul(a, b)
    elif isinstance(a, np.ndarray):
        _check_all_np_ndarray([b])
        return np.matmul(a, b)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def to_index_array(array):
    """Returns an array that is suitable for indexing a dimension of
    a different array.
    After a few checks (int, 1D), this operation will convert the dtype to
    torch.long (which is, in some torch versions, the only acceptable type
    of index tensor). Numpy arrays are left unchanged.
    """
    if len(array.shape) != 1:
        raise ValueError("Index arrays must be 1D")

    if isinstance(array, TorchTensor):
        if torch.is_floating_point(array):
            raise ValueError("Index arrays must be integers")
        return array.to(torch.long)
    elif isinstance(array, np.ndarray):
        if not np.issubdtype(array.dtype, np.integer):
            raise ValueError("Index arrays must be integers")
        return array
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def unique(array, axis: Optional[int] = None):
    """Find the unique elements of an array."""
    if isinstance(array, TorchTensor):
        return torch.unique(array, dim=axis)
    elif isinstance(array, np.ndarray):
        return np.unique(array, axis=axis)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def zeros_like(array, shape: Optional[List[int]] = None, requires_grad: bool = False):
    """
    Create an array filled with zeros, with the given ``shape``, and similar dtype,
    device and other options as ``array``.

    If ``shape`` is ``None``, the array shape is used instead. ``requires_grad`` is only
    used for torch tensors, and set the corresponding value on the returned array.

    This is the equivalent to ``np.zeros_like(array, shape=shape)``.
    """
    if isinstance(array, TorchTensor):
        if shape is None:
            shape = array.size()

        return torch.zeros(
            shape,
            dtype=array.dtype,
            layout=array.layout,
            device=array.device,
        ).requires_grad_(requires_grad)
    elif isinstance(array, np.ndarray):
        if shape is None:
            shape = array.shape
        return np.zeros_like(array, shape=shape, subok=False)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def where(array):
    """Return the indices where `array` is True.

    This function has the same behavior as ``np.where(array)``.
    """
    if isinstance(array, TorchTensor):
        return torch.where(array)
    elif isinstance(array, np.ndarray):
        return np.where(array)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def abs(array):
    """
    Returns the absolute value of the elements in the array.

    It is equivalent of np.abs(array) and torch.abs(tensor)
    """
    if isinstance(array, TorchTensor):
        return torch.abs(array)
    elif isinstance(array, np.ndarray):
        return np.abs(array).astype(array.dtype)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def argsort(array):
    """
    Returns the sorted arguments of the elements in the array.

    It is equivalent of np.argsort(array) and torch.argsort(tensor)
    """
    if isinstance(array, TorchTensor):
        return torch.argsort(array)
    elif isinstance(array, np.ndarray):
        return np.argsort(array)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def contiguous(array):
    """
    Returns a contiguous array.

    It is equivalent of np.ascontiguousarray(array) and tensor.contiguous(). In
    the case of numpy, C order is used for consistency with torch. As such, only
    C-contiguity is checked.
    """
    if isinstance(array, TorchTensor):
        if array.is_contiguous():
            return array
        return array.contiguous()
    elif isinstance(array, np.ndarray):
        if array.flags["C_CONTIGUOUS"]:
            return array
        return np.ascontiguousarray(array)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def to_int_list(array) -> List[int]:
    if isinstance(array, TorchTensor):
        # we need to do it this way because of
        # https://github.com/pytorch/pytorch/issues/76295
        return array.to(dtype=torch.int64).tolist()
    elif isinstance(array, np.ndarray):
        return array.tolist()
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def int_range_like(min_val: int, max_val: int, like):
    """
    Returns an array of integers from min to max, non-inclusive, based on the type of
    `like`

    It is equivalent of np.arange(start, end) and torch.arange(start, end) for the given
    array dtype and device.
    """
    if isinstance(like, TorchTensor):
        return torch.arange(min_val, max_val, dtype=like.dtype, device=like.device)
    elif isinstance(like, np.ndarray):
        return np.arange(min_val, max_val).astype(like.dtype)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def int_array_like(int_list: Union[List[int], List[List[int]]], like):
    """
    Converts the input list of int to a numpy array or torch tensor based on the type of
    ``like``. The resulting array/tensor will have the same dtype and device as
    ``like``.
    """
    if isinstance(like, TorchTensor):
        if torch.jit.isinstance(int_list, List[int]):
            return torch.tensor(int_list, dtype=like.dtype, device=like.device)
        else:
            return torch.tensor(int_list, dtype=like.dtype, device=like.device)
    elif isinstance(like, np.ndarray):
        return np.array(int_list).astype(like.dtype)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def real_array_like(float_list: List[float], like):
    """
    Converts the input list of float to a numpy array or torch tensor based on the array
    type of ``like``. The resulting array/tensor will have the same dtype and device as
    ``like``.
    """
    if isinstance(like, TorchTensor):
        return torch.tensor(float_list, dtype=like.dtype, device=like.device)
    elif isinstance(like, np.ndarray):
        return np.array(float_list).astype(like.dtype)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def bool_array_like(bool_list: List[bool], like):
    """
    Converts the input list of bool to a numpy array or torch tensor based on the type
    of ``like``. The resulting array/tensor will have the same device as ``like``.
    """
    if isinstance(like, TorchTensor):
        return torch.tensor(bool_list, dtype=torch.bool, device=like.device)
    elif isinstance(like, np.ndarray):
        return np.array(bool_list).astype(bool)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def cartesian_prod(array1, array2):
    """
    Imitates like itertools.product(array1, array2)
    """
    if isinstance(array1, TorchTensor) and isinstance(array2, TorchTensor):
        return torch.cartesian_prod(array1, array2)
    elif isinstance(array1, np.ndarray) and isinstance(array2, np.ndarray):
        # using itertools should be fastest way according to
        # https://stackoverflow.com/a/28684982
        return np.array(list(itertools.product(array1, array2)))
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def all(array, axis: Optional[int] = None):
    """Test whether all array elements along a given axis evaluate to True.

    This function has the same behavior as
    ``np.all(array,axis=axis)``.
    """
    if isinstance(array, bool):
        array = np.array(array)
    if isinstance(array, list):
        array = np.array(array)

    if isinstance(array, TorchTensor):
        # torch.all has two implementation, and picks one depending if more than one
        # parameter is given. The second one does not supports setting dim to `None`
        if axis is None:
            return torch.all(input=array)
        else:
            return torch.all(input=array, dim=axis)
    elif isinstance(array, np.ndarray):
        return np.all(a=array, axis=axis)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def max(array):
    """
    Takes the maximun value of the array.

    This function has the same behavior as
    ``np.max(array)`` or ``torch.max(array)``.
    """
    if isinstance(array, TorchTensor):
        return torch.max(array)
    elif isinstance(array, np.ndarray):
        return np.max(array)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def any(array):
    """Test whether any array elements along a given axis evaluate to True.

    This function has the same behavior as
    ``np.any(array)``.
    """
    if isinstance(array, bool):
        array = np.array(array)
    if isinstance(array, list):
        array = np.array(array)
    if isinstance(array, TorchTensor):
        return torch.any(array)
    elif isinstance(array, np.ndarray):
        return np.any(array)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def allclose(
    a: TorchTensor,
    b: TorchTensor,
    rtol: float,
    atol: float,
    equal_nan: bool = False,
):
    """Compare two arrays using ``allclose``

    This function has the same behavior as
    ``np.allclose(array1, array2, rtol, atol, equal_nan)``.
    """
    if isinstance(a, TorchTensor):
        _check_all_torch_tensor([b])
        return torch.allclose(
            input=a, other=b, rtol=rtol, atol=atol, equal_nan=equal_nan
        )
    elif isinstance(a, np.ndarray):
        _check_all_np_ndarray([b])
        return np.allclose(a=a, b=b, rtol=rtol, atol=atol, equal_nan=equal_nan)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def swapaxes(array, axis0: int, axis1: int):
    """Swaps axes of an array."""
    if isinstance(array, TorchTensor):
        return torch.swapaxes(array, axis0, axis1)
    elif isinstance(array, np.ndarray):
        return np.swapaxes(array, axis0, axis1)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def conjugate(array):
    """
    Conjugate the array

    This function has the same behavior as
    ``np.conjugate(array)`` or ``torch.conj(array)``.
    """
    if isinstance(array, TorchTensor):
        return torch.conj(array)
    elif isinstance(array, np.ndarray):
        return np.conjugate(array)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def real(array):
    """
    Takes the real part of the array

    This function has the same behavior as
    ``np.real(array)`` or ``torch.real(array)``.
    """
    if isinstance(array, TorchTensor):
        return torch.real(array)
    elif isinstance(array, np.ndarray):
        return np.real(array)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def imag(array):
    """
    Takes the imag part of the array

    This function has the same behavior as ``np.imag(array)`` or ``torch.imag(array)``.
    """
    if isinstance(array, TorchTensor):
        return torch.imag(array)
    elif isinstance(array, np.ndarray):
        return np.imag(array)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def roll(array, shifts: List[int], axis: List[int]):
    """
    Roll array elements along a given axis.

    This function has the same behavior as ``np.roll(array)`` or ``torch.roll(array)``.
    """
    if isinstance(array, TorchTensor):
        return torch.roll(array, shifts=shifts, dims=axis)
    elif isinstance(array, np.ndarray):
        return np.roll(array, shift=shifts, axis=axis)
    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)


def to(
    array,
    backend: Optional[str] = None,
    dtype: Optional[DType] = None,
    device: Optional[Union[str, Device]] = None,
):
    """Convert the array to the specified backend, dtype, and device"""
    if isinstance(array, TorchTensor):
        if backend is None:
            backend = "torch"

        if dtype is None:
            dtype = array.dtype
        if device is None:
            device = array.device
        if isinstance(device, str):
            device = torch.device(device)

        if backend == "torch":
            return array.to(dtype=dtype).to(device=device)

        elif backend == "numpy":
            if torch_jit_is_scripting():
                raise ValueError("cannot call numpy conversion when torch-scripting")
            else:
                return array.detach().cpu().numpy()

        else:
            raise ValueError(f"Unknown backend: {backend}")

    elif isinstance(array, np.ndarray):
        if backend is None:
            backend = "numpy"

        if backend == "numpy":
            return np.array(array, dtype=dtype)

        elif backend == "torch":
            return torch.tensor(array, dtype=dtype, device=device)
        else:
            raise ValueError(f"Unknown backend: {backend}")

    else:
        raise TypeError(UNKNOWN_ARRAY_TYPE)



import warnings

import numpy as np

from .._c_api import c_uintptr_t
from .base import SystemBase


try:
    import chemfiles

    if not chemfiles.__version__.startswith("0.10"):
        warnings.warn(
            message="found chemfiles, but the version is not supported: "
            "we need chemfiles v0.10.",
            category=ImportWarning,
            stacklevel=1,
        )
        HAVE_CHEMFILES = False
    else:
        HAVE_CHEMFILES = True
except ImportError:
    HAVE_CHEMFILES = False


# global cache of types number for atoms outside of the periodic table
ATOMIC_TYPES_CACHE = {}


def get_type_for_non_element(name):
    """Get type associated with atom that is not in the periodic table."""
    if name in ATOMIC_TYPES_CACHE:
        return ATOMIC_TYPES_CACHE[name]
    else:
        # start at 120 since that more atoms that are in the periodic table
        atomic_type = 120 + len(ATOMIC_TYPES_CACHE)
        ATOMIC_TYPES_CACHE[name] = atomic_type
        return atomic_type


class ChemfilesSystem(SystemBase):
    """Implements :py:class:`featomic.SystemBase` wrapping a `chemfiles.Frame`_.

    Since chemfiles does not offer a neighbors list, this implementation of system can
    only be used with ``use_native_system=True`` in
    :py:func:`featomic.calculators.CalculatorBase.compute`.

    Atomic type are assigned as the atomic number if the atom ``type`` is one of the
    periodic table elements; or as a value above 120 if the atom type is not in the
    periodic table.

    .. _chemfiles.Frame: http://chemfiles.org/chemfiles.py/latest/reference/frame.html
    """

    @staticmethod
    def can_wrap(o):
        return isinstance(o, chemfiles.Frame)

    def __init__(self, frame):
        """
        :param frame : `chemfiles.Frame`_ object object to be wrapped
            in this ``ChemfilesSystem``
        """
        super().__init__()
        if not isinstance(frame, chemfiles.Frame):
            raise Exception("this class expects chemfiles.Frame objects")

        self._frame = frame
        self._type = np.zeros((self.size()), dtype=c_uintptr_t)
        for i, atom in enumerate(self._frame.atoms):
            if atom.atomic_number != 0:
                self._type[i] = atom.atomic_number
            else:
                self._type[i] = get_type_for_non_element(atom.type)

    def size(self):
        return len(self._frame.atoms)

    def types(self):
        return self._type

    def positions(self):
        return self._frame.positions

    def cell(self):
        # we need to transpose the cell since chemfiles stores it in column
        # major format
        return self._frame.cell.matrix.T

    def compute_neighbors(self, cutoff):
        raise Exception(
            "chemfiles systems can only be used with 'use_native_system=True'"
        )

    def pairs(self):
        raise Exception(
            "chemfiles systems can only be used with 'use_native_system=True'"
        )

    def pairs_containing(self, atom):
        raise Exception(
            "chemfiles systems can only be used with 'use_native_system=True'"
        )
